#!/usr/bin/python3

from math import pi, sqrt
import unittest
import warnings
import os
import filecmp

import numpy as np

import hubbard.substrate.substrate
import hubbard.substrate.dftsubstrate
import hubbard.substrate.vaspband2np

class TestBaseSubstrate(unittest.TestCase):
    def test_init(self):
        sub = hubbard.substrate.substrate.BaseSubstrate()
        msg = " did not have the expected value."
        self.assertEqual(sub.nx, 1, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 1, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0]]), msg="glist"+msg)
    #
    def test_set_params(self):
        sub = hubbard.substrate.substrate.BaseSubstrate()
        msg = " did not have the expected value."
        sub.set_params(nx=1, ny=1)
        self.assertEqual(sub.nx, 1, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 1, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0]]), msg="glist"+msg)
        sub.set_params(nx=2)
        self.assertEqual(sub.nx, 2, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 2, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0],[1,0]]), msg="glist"+msg)
        sub.set_params(nx=3, ny=2)
        self.assertEqual(sub.nx, 3, msg="nx"+msg)
        self.assertEqual(sub.ny, 2, msg="ny"+msg)
        self.assertEqual(sub.ncells, 6, msg="ncells"+msg)
        self.assertEqual(len(sub.glist), 6, msg="len(glist)"+msg)
        # The order of points is not important, only that they are all present.
        glist2 = [[0,0], [1,0], [2,0], [0,1], [1,1], [2,1]]
        self.assertTrue(all([any([np.all(k == kp) for k in sub.glist]) for kp in glist2]),
                        msg="glist"+msg)
    #
    def test_get_coupling(self):
        sub = hubbard.substrate.substrate.BaseSubstrate(nx=2, ny=1)
        # glist = [[0,0], [1,0]]
        def msg(k,pos,couple):
            return "k="+str(k)+", pos="+str(pos)+", couple="+str(couple)+", get_coupling did not produce expected array."
        # k = [0,0]; positions = [[0,0]]
        k = [0,0]
        pos = [[0,0]]
        expected = -1/sqrt(2)*np.array([[1,1]])
        arr = sub.get_coupling(k, pos)
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))
        # Do multiple sites
        pos = np.zeros((3,2))
        arr = sub.get_coupling(k, pos)
        expected = -1/sqrt(2)*np.ones((3,2))
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))
        # Do unusual coupling
        couple = [0.5, 1, 2]
        arr = sub.get_coupling(k, pos, couple)
        expected = -1/sqrt(2) * np.array([[0.5,0.5], [1,1], [2,2]])
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,couple))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,couple))
        # Do non-zero pos
        pos = [[0.5, 0.5]]
        arr = sub.get_coupling(k, pos)
        expected = -1/sqrt(2) * np.array([[1, -1]])
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))
        # Non-zero k
        k = [0,0.5]
        arr = sub.get_coupling(k, pos)
        expected = -1/sqrt(2) * np.array([[1j, -1j]])
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))
        # Do a couple of positions
        pos = [[0,0], [0.5,0.5]]
        k = [0,0]
        arr = sub.get_coupling(k, pos)
        expected = -1/sqrt(2) * np.array([[1,1], [1,-1]])
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))
        k = [0, 0.5]
        arr = sub.get_coupling(k, pos)
        expected = -1/sqrt(2) * np.array([[1,1], [1j,-1j]])
        self.assertEqual(expected.shape, arr.shape, msg=msg(k,pos,1))
        self.assertAlmostEqual(np.abs(expected - arr).sum(), 0, msg=msg(k,pos,1))

class TestSquareSubstrate(unittest.TestCase):
    mycls = hubbard.substrate.substrate.SquareSubstrate
    def en(self,k):
        """Energy dispersion"""
        return -2 * (np.cos(2*pi*k[0]) + np.cos(2*pi*k[1]))
    #
    def test_init(self):
        sub = self.mycls()
        msg = " did not have the expected value."
        self.assertEqual(sub.nx, 1, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 1, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0]]), msg="glist"+msg)
        self.assertEqual(sub.t, 1, msg="t"+msg)
        self.assertEqual(sub.offset, 0, msg="offset"+msg)
    #
    def test_set_params(self):
        sub = self.mycls()
        msg = " did not have the expected value."
        sub.set_params(nx=1, ny=1, t=2, offset=-1)
        self.assertEqual(sub.nx, 1, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 1, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0]]), msg="glist"+msg)
        self.assertEqual(sub.t, 2, msg="t"+msg)
        self.assertEqual(sub.offset, -1, msg="offset"+msg)
        sub.set_params(nx=2)
        self.assertEqual(sub.nx, 2, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 2, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0],[1,0]]), msg="glist"+msg)
        self.assertEqual(sub.t, 2, msg="t"+msg)
        self.assertEqual(sub.offset, -1, msg="offset"+msg)
        sub.set_params(t=3, offset=-0.5)
        self.assertEqual(sub.nx, 2, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 2, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0],[1,0]]), msg="glist"+msg)
        self.assertEqual(sub.t, 3, msg="t"+msg)
        self.assertEqual(sub.offset, -0.5, msg="offset"+msg)
    #
    def test_dispersion(self):
        sub = self.mycls()
        msg = "dispersion was not the expected value."
        # A few semi-arbitrary k-points to test.
        ks = [[0,0], [0.5,1], [-1,0.2]]
        for k in ks:
            self.assertAlmostEqual(self.en(k), sub.dispersion(k), msg=msg)
        # Change the parameters.
        sub.set_params(t=2, offset=-1)
        for k in ks:
            self.assertAlmostEqual(2*self.en(k)-1, sub.dispersion(k), msg=msg)
        sub.set_params(t=1, offset=0, nx=2, ny=3)
        for k in ks:
            self.assertAlmostEqual(self.en([k[0]/2,k[1]/3]), sub.dispersion(k), msg=msg)
    #
    def test_get_matrix(self):
        sub = self.mycls()
        msg1 = "matrix was not the expected size."
        msg2 = "matrix was not the expected value."
        # A few semi-arbitrary k-points to test.
        ks = [[0,0], [0.5,1], [-1,0.2]]
        for k in ks:
            mat = sub.get_matrix(k)
            expected = np.array([[self.en(k)]])
            self.assertEqual(mat.size, expected.size, msg=msg1)
            self.assertAlmostEqual(np.abs(mat-expected).sum(), 0, msg=msg2)
        # A larger cell.
        sub.set_params(nx=2, ny=1)
        for k in ks:
            mat = sub.get_matrix(k)
            expected = np.diag([self.en([k[0]/2,k[1]]), self.en([1/2+k[0]/2,k[1]])])
            self.assertEqual(mat.size, expected.size, msg=msg1)
            self.assertAlmostEqual(np.abs(mat-expected).sum(), 0, msg=msg2)

class TestTriangleSubstrate(TestSquareSubstrate):
    # TriangleSubstrate inherits almost everything from SquareSubstrate,
    # so I can inherit the tests as well.
    mycls = hubbard.substrate.substrate.TriangleSubstrate
    def en(self,k):
        """Energy dispersion"""
        return -2 * (np.cos(2*pi*k[0]) + np.cos(2*pi*k[1]) + np.cos(2*pi*(k[1]-k[0])))

class TestDFTSubstrate(unittest.TestCase):
    ddir = "fixtures/"
    def en(self, k):
        # Manually calculated bilinear interpolation for band1.dat
        k = np.mod(k,1)
        if k[0] < 1/3:
            x = 3*k[0]
        elif k[0] > 2/3:
            x = 3*(1-k[0])
        else:
            x = 1
        if k[1] < 1/2:
            y = 2*k[1]
        else:
            y = 2*(1-k[1])
        return x * y
    #
    def test_init(self):
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate()
        msg = " did not have the expected value."
        self.assertEqual(sub.nx, 1, msg="nx"+msg)
        self.assertEqual(sub.ny, 1, msg="ny"+msg)
        self.assertEqual(sub.ncells, 1, msg="ncells"+msg)
        self.assertTrue(np.all(sub.glist == [[0,0]]), msg="glist"+msg)
        self.assertEqual(sub.offset, 0, msg="offset"+msg)
    #
    def test_loadarray(self):
        data = np.array([[0, 0], [0, 1], [0, 1]], dtype=float)
        msg = " did not have expected value."
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate()
        sub.loadarray(data, gamma=True)
        self.assertAlmostEqual(np.abs(data - sub.datagrid).sum(), 0,
                               msg="datagrid"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0,0, grid=False)), 0,
                               msg="interpolated at (0,0)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0.5,0.5,grid=False)), 1,
                               msg="interpolated at (0.5,0.5)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1,1,grid=False)), 0,
                               msg="interpolated at (1,1)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1/3,0.5,grid=False)), 1,
                               msg="interpolated at (1/3,0.5)"+msg)
    #
    def test_loadarray_notgamma(self):
        data = np.array([[0, 0], [0, 1], [0, 1]], dtype=float)
        msg = " did not have expected value."
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate()
        sub.loadarray(data, gamma=False)
        self.assertAlmostEqual(np.abs(data - sub.datagrid).sum(), 0,
                               msg="datagrid"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0,1/4, grid=False)), 0,
                               msg="interpolated at (0,1/4)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0.5,3/4,grid=False)), 1,
                               msg="interpolated at (0.5,3/4)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1,1/4,grid=False)), 0,
                               msg="interpolated at (1,1/4)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1/3,3/4,grid=False)), 1,
                               msg="interpolated at (1/3,0.5)"+msg)
    #
    def test_loadtxt(self):
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate()
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            # If it's raising warnings, something's gone wrong.
            sub.loadtxt(self.ddir+"band1.dat")
        data = np.array([[0, 0], [0, 1], [0, 1]], dtype=float)
        msg = " did not have expected value."
        self.assertAlmostEqual(np.abs(data - sub.datagrid).sum(), 0,
                               msg="datagrid"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0,0, grid=False)), 0,
                               msg="interpolated at (0,0)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(0.5,0.5,grid=False)), 1,
                               msg="interpolated at (0.5,0.5)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1,1,grid=False)), 0,
                               msg="interpolated at (1,1)"+msg)
        self.assertAlmostEqual(float(sub.interpolated(1/3,0.5,grid=False)), 1,
                               msg="interpolated at (1/3,0.5)"+msg)
    #
    def test_dispersion(self):
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate(filename=self.ddir+"band1.dat")
        msg = "dispersion was not the expected value."
        # A few semi-arbitrary k-points to test.
        ks = [[0,0], [0.5,1], [-1,0.2]]
        for k in ks:
            self.assertAlmostEqual(self.en(k), sub.dispersion(k), msg=msg)
        # Change the parameters.
        sub.set_params(offset=-1)
        for k in ks:
            self.assertAlmostEqual(self.en(k)-1, sub.dispersion(k), msg=msg)
        sub.set_params(offset=0, nx=2, ny=3)
        for k in ks:
            self.assertAlmostEqual(self.en([k[0]/2,k[1]/3]), sub.dispersion(k), msg=msg)
    #
    def test_get_matrix(self):
        sub = hubbard.substrate.dftsubstrate.DFTSubstrate(filename=self.ddir+"band1.dat")
        msg1 = "matrix was not the expected size."
        msg2 = "matrix was not the expected value."
        # A few semi-arbitrary k-points to test.
        ks = [[0,0], [0.5,1], [-1,0.2]]
        for k in ks:
            mat = sub.get_matrix(k)
            expected = np.array([[self.en(k)]])
            self.assertEqual(mat.size, expected.size, msg=msg1)
            self.assertAlmostEqual(np.abs(mat-expected).sum(), 0, msg=msg2)
        # A larger cell.
        sub.set_params(nx=2, ny=1)
        for k in ks:
            mat = sub.get_matrix(k)
            expected = np.diag([self.en([k[0]/2,k[1]]), self.en([1/2+k[0]/2,k[1]])])
            self.assertEqual(mat.size, expected.size, msg=msg1)
            self.assertAlmostEqual(np.abs(mat-expected).sum(), 0, msg=msg2)

class Testvaspband2np(unittest.TestCase):
    ddir = "fixtures/"
    def test_read_eigenval(self):
        """Check that read_eigenval evaluates without errors."""
        eigen = hubbard.substrate.vaspband2np.load_eigenval(self.ddir+'EIGENVAL.dos',outcar=self.ddir+'OUTCAR.dos')
        nkpt = 217
        nband = 24
        self.assertEqual(eigen.bands.shape, (nkpt,nband), msg='EIGENVAL bands array not the right shape')
        self.assertEqual(eigen.kpoints.shape, (nkpt,3), msg='EIGENVAL kpoints array not the right shape')
    #
    def test_savetxt(self):
        """Check that savetxt_band gives the expected result."""
        tmpfile = self.ddir+'tmp'
        # Delete the temporary file if it exists
        try:
            os.remove(tmpfile)
        except FileNotFoundError:
            pass
        # Load PROCAR
        procar = hubbard.substrate.vaspband2np.load_procar(self.ddir+'PROCAR.dos',outcar=self.ddir+'OUTCAR.dos')
        # N.B. PROCAR and EIGENVAL have different precisions for the same data.
        hubbard.substrate.vaspband2np.savetxt_band(tmpfile, procar, 16, self.ddir+'OUTCAR.dos')
        # Check if it matches existing file.
        msg='Generated file did not match expected file. diff '+tmpfile+' and Ag_band17.dat for comparison.'
        self.assertTrue(filecmp.cmp(tmpfile, self.ddir+'Ag_band17.dat', shallow=False), msg=msg)
        # If we get to the end in one piece, remove the tmpfile
        os.remove(tmpfile)



if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3

"""
Base class for Mean Field Hubbard model with Brillouin zone sampling.
Has the mixing routines and other things which aren't dependent on the
specifics of the lattice.

The primary change from the Gamma point only code is in the eigenstep
method, which has been rebuilt to allow k-point sampling.
The kinetic energy matrix is also no longer stored but must be
dynamically constructed for each k-point.

Several things must be defined in the subclasses.
 - get_kinetic and set_kinetic. get_kinetic needs to be callable after
    initialisation. THIS ONE IS IMPORTANT.
 - The copy method. I don't know which data you need to feed into new instances.
 - get_coordinates - returns a (nsites, 2) ndarray which indicates
     the Cartesian coordinates of each site. Used for plotting.
     (I assume a 2D lattice for plotting. 3D plotting will require re-working
     the plotting routines to do 3D plots. That's a job for a subclass.)
 - save and load methods (you don't have to define them, but they aren't
     defined here). (I may find a way to abstract the saving/loading later.)
 - any bespoke electron density setting methods, in
     _electron_density_single_methods.
There is also the variable self.dims. It tells you how many dimensions the
k-space vectors exist in. It defaults to 2, but you can make it any positive
integer. Please don't change it after initialisation.

Created: 2020-09-16
Last Modified: 2022-11-03
Author: Bernard Field
    Copyright (C) 2021 Bernard Field, GNU GPL v3+
"""

from warnings import warn

import numpy as np
import matplotlib.pyplot as plt

from hubbard.misc import ConvergenceWarning, MixingError, fermi_distribution,\
     random_density, choose_alpha, bounded_random_numbers_with_sum_dirichlet,\
     random_points_density

class HubbardKPoints():
    """Base class for Mean Field Hubbard model with Brillouin zone sampling."""
    #
    ## IO
    #
    def __init__(self,nsites,u=0,nup=0,ndown=0,allow_fractions=False,**kwargs):
        """
        Initialises things.
        Crucially, this initialiser does NOT set the kinetic energy
        part of the Hamiltonian. That is lattice specific.

        Inputs: nsites - positive integer. Number of sites. Cannot be
                changed later.
            u - number. Hubbard U parameter.
            nup,ndown,kwargs - arguments for set_electrons.
        Last Modified: 2020-10-07
        """
        if nsites <= 0:
            raise ValueError("Number of sites must be positive.")
        self.nsites = nsites
        """Number of sites."""
        self.u = u
        """Hubbard U parameter."""
        self.allow_fractions = allow_fractions
        """Flag for allow_fractions (toggle with toggle_allow_fractions)."""
        self.mag = 0
        """Magnetic field."""
        self.set_electrons(nup,ndown,**kwargs,backup=False)
        # If number of dimensions has not already been set, set it.
        if not hasattr(self, 'dims'):
            self.dims = 2
            """Number of dimensions in momentum-space."""
        if not hasattr(self, 'reclat'):
            self.reclat = np.eye(self.dims)
            """Reciprocal lattice vectors. Used for plotting."""
        # Initialise the k-mesh as a single k-point.
        self.set_kmesh(*([1]*self.dims))
    #
    def copy(self):
        """
        MUST BE OVERRIDDEN IN CHILD CLASS.
        This method is meant to return a copy of the current instance.

        Expected Output: Hubbard-derived object.
        """
        raise NotImplementedError
    #
    @classmethod
    def load(cls,f):
        """
        MUST BE OVERRIDDEN IN CHILD CLASS.
        Loads a Hubbard-derived object from file f.
        Inputs: f - filename.
        Outputs: Hubbard-derived object
        """
        raise NotImplementedError
    #
    def save(self,f):
        """
        MUST BE OVERRIDDEN IN CHILD CLASS.
        Save this object to file f.
        Inputs: f - filename
        Writes a file to f.
        """
        raise NotImplementedError
    #
    ## ELECTRON DENSITY MODIFIERS
    #
    def anderson_mixing(self,max_iter=100,rdiff=1e-6,mix=0.5,T=None,mu=None):
        """
        A more elaborate scheme than linear mixing. It adds a sort of
        extra internal linear mixing step with an automatically chosen
        ratio and including information from the previous step.
        Here I minimise the residual.
        See Sec IIB of D.D.Johnson 1988 PRB v38 n 18 and
        Ref's 3 and 4 therein for the algorithm.

        Inputs: max_iter - positive integer, maximum number of iterations.
            rdiff - postivie number, value of the residual at which we should
                stop the cycle.
            mix - number between 0 and 1. Mixing parameter.
            T - optional, non-negative number. Temperature for determining
                occupation of states. If provided, allows fractional occupation
                and exchange between spin up and down channels.
            mu - optional, number. Chemical potential for Grand Canonical
                Ensemble. Number of electrons becomes non-constant. Requires
                fractional electrons and setting T.
        Output: number, the residual
        Last Modified: 2021-04-16
        """
        # Check if we have specified a temperature
        if T is not None:
            fermi = True
            if not self.allow_fractions:
                warn("allow_fractions is currently False. Mixing with "
                     "finite T requires this flag to be enabled. Setting "
                     "allow_fractions to True.")
                self.toggle_allow_fractions(True)
        else:
            fermi = False
        if mu is not None:
            if not fermi:
                raise TypeError("If mu is provided, T must be provided.")
        # Set Grand Canonical Ensemble flag
        gce = mu is not None
        # We have to run an initial step by normal linear mixing.
        if fermi:
            _, nupoutold, ndownoutold = self._eigenstep_finite_T(T)
        else:
            _, nupoutold, ndownoutold = self._eigenstep()
        # Residues
        resold = np.concatenate((nupoutold - self.nup,ndownoutold - self.ndown))
        # If the residue is sufficiently small, we're done.
        if np.linalg.norm(resold) < rdiff:
            return np.linalg.norm(resold)
        # Record old input densities.
        nupinold = self.nup
        ndowninold = self.ndown
        # Do mixing
        self.set_electrons(nup=self.nup*(1-mix)+ nupoutold*mix,
                           ndown=self.ndown*(1-mix) + ndownoutold*mix)
        for i in range(max_iter):
            # Put in electron density, get electron density out.
            if fermi:
                if gce:
                    _, nupout, ndownout = self._eigenstep_GCE(T,mu)
                else:
                    _, nupout, ndownout = self._eigenstep_finite_T(T)
            else:
                _, nupout, ndownout = self._eigenstep()
            # Get residue
            res = np.concatenate((nupout-self.nup,ndownout-self.ndown))
            # If the residue is sufficiently small, we're done.
            if np.linalg.norm(res) < rdiff:
                return np.linalg.norm(res)
            # Get Anderson mixing parameter
            beta = np.dot(res,res-resold)/(np.dot(res-resold,res-resold))
            # Do the Anderson pre-mixing mixing
            nupinnew = (1-beta)*self.nup + beta*nupinold
            nupoutnew = (1-beta)*nupout + beta*nupoutold
            ndowninnew = (1-beta)*self.ndown + beta*ndowninold
            ndownoutnew = (1-beta)*ndownout + beta*ndownoutold
            #Record input densities.
            nupinold = self.nup
            ndowninold = self.ndown
            # Do main mixing
            nup = nupinnew*(1-mix) + nupoutnew*mix
            ndown = ndowninnew*(1-mix) + ndownoutnew*mix
            # Check validity.
            if nup.max()>1 or nup.min()<0 or ndown.max()>1 or ndown.min()<0:
                # Electron density is out of bounds.
                # Fall back on simple linear mixing.
                nup = self.nup*(1-mix) + nupout*mix
                ndown = self.ndown*(1-mix) + ndownout*mix
            # Apply mixing
            self.set_electrons(nup=nup,ndown=ndown)
            # Record old output densities
            nupoutold = nupout
            ndownoutold = ndownout
            # Record old residue
            resold = res
        # Done, but due to exceeding the maximum iterations.
        warn("Self-consistency not reached after "+str(max_iter)+" steps. "
             "Residual in last step is "+str(np.linalg.norm(res))+".")
        return np.linalg.norm(res)
    #
    def linear_mixing(self,max_iter=100,ediff=1e-6,mix=0.5,rdiff=5e-4,T=None,print_residual=False,mu=None,return_energy=True,debug=False):
        """
        Iterates the Hubbard Hamiltonian towards self-consistency.
        Uses linear density mixing. pnew = (1-mix)*pin + mix*pout

        Inputs: mix - number between 0 and 1. Mixing factor. Default 0.5
            ediff - number. When energy step is less than this, stop.
                Default 1e-6.
            max_iter - integer. Max number of iterations. Default 100.
            rdiff - number. When residual is less than this, stop.
                Both ediff and rdiff must be satisfied.
            T - optional positive number. If set, we fill states using the
                Fermi-Dirac distribution rather than just filling the N
                lowest eigenstates. If set, the number of up and down
                spins is likely to vary, although the total number of
                spins should be constant. Requires allow_fractions=True.
            print_residual - Boolean, default False. Prints the residual
                of the density. You want this to be small.
                Set False if you are looping through stuff to avoid flooding
                stdout.
            mu - optional number. Chemical potential. If set, calculates mixing
                in the grand canonical ensemble. Requires T and 
                allow_fractions=True. Number of electrons will vary.
            return_energy - Boolean. If true, return the energy. If false,
                return the residual.
            debug - Boolean. If True, records the density and other variables
                at each step and returns it in a dictionary.
        Outputs: if not debug: a number- energy or residual
            if debug: a dictionary of lists. Each list entry is the value at
            the end of each step. (Except 'energy', which is the value at
            the start of the step.) Dictionary has nup, ndown, residual, and
            energy.
        Effects: sets nup and ndown to new values.

        Last Modified: 2021-04-16
        """
        # Check if we have specified a temperature
        if T is not None:
            fermi = True
            if not self.allow_fractions:
                warn("allow_fractions is currently False. Mixing with "
                     "finite T requires this flag to be enabled. Setting "
                     "allow_fractions to True.")
                self.toggle_allow_fractions(True)
        else:
            fermi = False
        if mu is not None:
            if not fermi:
                raise TypeError("If mu is provided, T must be provided.")
        # Set Grand Canonical Ensemble flag
        gce = mu is not None
        # Initialise debug data
        if debug:
            debugdata = {'nup':[], 'ndown':[], 'residual':[], 'energy':[]}
        # Initialise first energy as a very large number.
        en = 1e12
        for i in range(max_iter):
            # Do a step
            if fermi:
                if gce:
                    ennew,nupnew,ndownnew = self._eigenstep_GCE(T,mu)
                else:
                    ennew,nupnew,ndownnew = self._eigenstep_finite_T(T)
            else:
                ennew,nupnew,ndownnew = self._eigenstep()
            # Do mixing
            residual_up = nupnew - self.nup
            residual_down = ndownnew - self.ndown
            # While analytically it should be impossible for the values to fall
            # outside the allowed range, floating point errors can cause it
            # to be outside by a tiniest smigeon. Catch these.
            nupnext = self.nup*(1-mix) + nupnew*mix
            ndownnext = self.ndown*(1-mix) + ndownnew*mix
            nupnext[nupnext > 1] = 1
            nupnext[nupnext < 0] = 0
            ndownnext[ndownnext > 1] = 1
            ndownnext[ndownnext < 0] = 0
            self.set_electrons(nup=nupnext, ndown=ndownnext)
            # Check for convergence
            res = np.linalg.norm(np.concatenate((residual_up,residual_down)))
            de = abs(ennew - en)
            # Before breaking, record debug data.
            if debug:
                debugdata['nup'].append(nupnext)
                debugdata['ndown'].append(ndownnext)
                debugdata['residual'].append(res)
                debugdata['energy'].append(ennew)
            # Check for convergence.
            if de < ediff and res < rdiff:
                break
            en = ennew
        # Raise warning if needed
        if de >= ediff or res >= rdiff:
            warn("Self-consistency not reached after "+str(max_iter)+" steps. "
                 "Residual = "+str(res)+", Ediff = "+str(de)+".",
                 ConvergenceWarning)
        if print_residual:
            print("Residual = "+str(res)+", Ediff = "+str(de))
        if debug:
            return debugdata
        elif return_energy:
            return en
        else:
            return res
    #
    def pulay_mixing(self,max_iter=100,rdiff=1e-3,T=None, mu=None, debug=False):
        """
        Performs Direct Inversion of the Iterative Subspace (aka Pulay mixing)
        to find the self-consistent electron density. Linear mixing is
        used for the simple relaxation steps (needed to expand the subspace).
        Should only be used when relatively close to convergence.
        Pulay 1980 Chemical Physics Letters vol 73 is 2 p 393-398
        
        Inputs: max_iter - positive integer, maximum number of iterations.
            rdiff - postivie number, value of the residual at which we should
                stop the cycle.
            T - optional, non-negative number. Temperature for determining
                occupation of states. If provided, allows fractional occupation
                and exchange between spin up and down channels.
            mu - optional number. Chemical potential. If set, calculates mixing
                in the grand canonical ensemble. Requires T and 
                allow_fractions=True. Number of electrons will vary.
                NOTE: Pulay mixing is very slow at converging the total number
                of electrons. If using mu, either stick with linear mixing or 
                use a much stricter initial convergence threshold.
            debug - Boolean. If True, records the density and other variables
                at each step and returns it in a dictionary.
        Outputs: if not debug: the residual
            if debug: a dictionary of lists. Each list entry is the value at
            the end of each step. Dictionary has nup, ndown, and residual.
        Output: number, the residual
        Last Modified: 2021-05-05
        """
        # Check if T is specified.
        if T is not None:
            fermi = True
            if not self.allow_fractions:
                warn("allow_fractions is currently False. Mixing with "
                     "finite T requires this flag to be enabled. Setting "
                     "allow_fractions to True.")
                self.toggle_allow_fractions(True)
        else:
            fermi = False
        if mu is not None:
            if not fermi:
                raise TypeError("If mu is set, T must be set.")
        # Set Grand Canonical Ensemble flag
        gce = mu is not None
        # Initialise debug record
        if debug:
            debugdata = {'nup':[], 'ndown':[], 'residual':[]}
        condmax = 1e4 # Maximum permissible value of the condition number.
        # (I haven't actually tested what a good value for this is,
        # but it seems to work okay most of the time.)
        nsites = self.nsites
        # Initialise our records.
        densities_up = np.empty((0,nsites))
        densities_down = np.empty((0,nsites))
        residues = np.empty((0,2*nsites))
        # Iterate
        for i in range(max_iter):
            # An eigenstep
            if fermi:
                if gce:
                    _, nup, ndown = self._eigenstep_GCE(T,mu)
                else:
                    _, nup, ndown = self._eigenstep_finite_T(T)
            else:
                _, nup, ndown = self._eigenstep()
            # Residue
            res = np.concatenate((nup - self.nup, ndown - self.ndown))
            # If the residue is sufficiently small, we're done.
            if np.linalg.norm(res) < rdiff:
                if debug:
                    debugdata['nup'].append(self.nup.copy())
                    debugdata['ndown'].append(self.ndown.copy())
                    debugdata['residual'].append(np.linalg.norm(res))
                    return debugdata
                else:
                    return np.linalg.norm(res)
            # Record
            densities_up = np.vstack((densities_up,nup))
            densities_down = np.vstack((densities_down,ndown))
            residues = np.vstack((residues,res))
            # Construct the error matrix
            if i==0:
                matrix = np.array([[np.dot(res,res)]])
            else:
                vect = np.dot(residues[0:-1],res)
                ee = np.dot(res,res)
                matrix = np.c_[np.vstack((matrix,vect)),np.r_[vect,ee]]
            # Set up the DIIS equation to solve, in Lagrange multiplier form.
            # Pulay 1980 Eq 6.
            diismat = -np.ones((len(matrix)+1,len(matrix)+1))
            diismat[-1,-1] = 0
            diismat[0:-1,0:-1] = matrix
            # Check the condition number. If too high, we'll need to omit
            # old entries to get back to linear independence.
            if np.linalg.cond(diismat) > condmax:
                diismat = diismat[1:,1:]
                densities_up = densities_up[1:]
                densities_down = densities_down[1:]
                residues = residues[1:]
                matrix = matrix[1:,1:]
            # The vector to equal.
            b = np.zeros(len(diismat))
            b[-1] = -1
            # Solve.
            coef = np.linalg.solve(diismat,b)
            # This vector holds the coefficients, and also at the end the
            # lagrange multiplier which should equal the norm of the residium
            # vector squared.
            # Write new densities.
            nupnew = np.dot(coef[0:-1],densities_up)
            ndownnew = np.dot(coef[0:-1],densities_down)
            # Floating point errors can take densities out of bounds in
            # limiting cases. Catch these cases.
            if self.nelectup == self.nsites:
                nupnew = np.ones(self.nsites)
            elif self.nelectup == 0:
                nupnew = np.zeros(self.nsites)
            if self.nelectdown == self.nsites:
                ndownnew = np.ones(self.nsites)
            elif self.nelectdown == 0:
                ndownnew = np.zeros(self.nsites)
            # Record debug data
            if debug:
                debugdata['nup'].append(nupnew)
                debugdata['ndown'].append(ndownnew)
                debugdata['residual'].append(np.linalg.norm(res))
            # Check validity
            if (nupnew.max()>1 or nupnew.min()<0
                or ndownnew.max()>1 or ndownnew.min()<0):
                if debug:
                    raise MixingError("Electron density out of bounds. Try linear_mixing.", debugdata)
                else:
                    raise MixingError("Electron density out of bounds. Try linear_mixing.")
            # Apply mixing.
            self.set_electrons(nupnew,ndownnew)
        # Done, but due to exceeding the maximum iterations.
        warn("Self-consistency not reached after "+str(max_iter)+" steps. "
             "Residual in last step is "+str(np.linalg.norm(res))+".",
             ConvergenceWarning)
        if debug:
            return debugdata
        else:
            return np.linalg.norm(res)
    #
    def set_electrons(self,nup=None,ndown=None,backup=True,**kwargs):
        """
        Sets the electron densities.

        Inputs: nup, ndown - either an integer for number of electrons
                or a list-like for the electron density.
                If integer, must be between 0 and number of sites.
                If list-like, must be of right length and have
                values between 0 and 1.
                May also be None, in which case it is not set.
            backup - Boolean, default True. Does backing up of the electron
                density in case of an error. Set to False if you are running
                as part of the initialisation.
        Keyword Arguments:
            method - string. {'uniform', 'random'}. Specifies
                how to generate the electron density if a number
                of electrons is provided. Ignored if n is a list.
            alpha - optional positive number. Used for random method.
                Dirichlet alpha parameter. Default is chosen automatically
                to be as small as practical.
            
        Effects: Sets nup, ndown, nelectup, nelectdown.
            In the event of an Exception, no changes are made.

        Last Modified: 2020-08-11
        """
        # Process keyword arguments. kwargs.get(key,default)
        method = kwargs.pop('method','uniform')
        alpha = kwargs.pop('alpha',None)
        # Any remaining kwargs are class specific.
        if backup:
            # Write backups
            nup_backup = self.nup
            nelectup_backup = self.nelectup
            ndown_backup = self.ndown
            nelectdown_backup = self.nelectdown
        try:
            # Set spin up if applicable.
            if nup is not None:
                self.nup,self.nelectup = self._electron_density_single(
                    nup,method,up=True,alpha=alpha,**kwargs)
            # Set spin down if applicable.
            if ndown is not None:
                self.ndown,self.nelectdown = self._electron_density_single(
                    ndown,method,up=False,alpha=alpha,**kwargs)
        except:
            if backup:
                # In the event of an error, undo all the changes.
                self.nup = nup_backup
                self.nelectup = nelectup_backup
                self.ndown = ndown_backup
                self.nelectdown = nelectdown_backup
            raise
    #
    def _density_from_vectors(self,evect,nelect):
        """
        Get electron density from eigenvectors and number of electrons.
        
        Do this separately for spin up and down.

        Inputs: evect - array, eigenvector output of np.linalg.eigh
            nelect - integer. Number of electrons occupying these states.
        Output: a 1D array of length of each eigenvector, of real non-negative
            numbers, representing the local electron density.

        Last Modified: 2020-09-16
        """
        nstates = nelect * self.kpoints
        # Take the absolute value squared of each eigenvector
        # then sum over the bottom nelect eigenvectors.
        if not self.allow_fractions:
            density = np.sum(abs(evect[:,0:nstates])**2,1)
        else:
            density = np.sum(abs(evect[:,0:int(nstates)])**2,1)
            # Get the fractional part of the occupations.
            frac = nstates - int(nstates)
            # Add energy from fractionally occupied states.
            density += abs(evect[:,int(nstates)])**2 * frac
        # Divide by kpoints to normalise.
        return density/self.kpoints
    #
    def _eigenstep(self):
        """
        Evaluate the Hamiltonian, get new energy and density.

        This is the core part of the self-consistency loop.
        Takes the electron densities and the parameters needed to
        make the Hamiltonian, then returns the energy and the
        electron densities predicted by the eigenvectors.
        Assumes the total number of up and down electrons stays
        constant.
        The idea is that you'll use the output densities to mix with
        the initial densities then use as new input densities in a
        self-consistency loop.

        Presets: nup, ndown, u, set_kinetic, nelectup, nelectdown
        Outputs: energy (real number), nup, ndown

        Last Modified: 2020-09-16
        """
        # Generate the potential energy.
        _, _, offset = self._potential()
        # Solve for the single-electron energy levels.
        eup, edown, vup, vdown = self._eigensystem()
        # Calculate the energy by filling electrons.
        energy = self._energy_from_states(eup,edown,offset)
        # Calculate the new densities.
        nupnew = self._density_from_vectors(vup,self.nelectup)
        ndownnew = self._density_from_vectors(vdown,self.nelectdown)
        # Return
        return energy, nupnew, ndownnew
    #
    def _eigenstep_finite_T(self,T):
        """
        Evaluate the Hamiltonian, get new energy and density. Use finite T.

        Takes the electron densities and the parameters needed to
        make the Hamiltonian, then returns the energy and the
        electron densities predicted by the eigenvectors.
        Fills the eigenstates by assuming a Fermi-Dirac distribution
        with temperature T.

        Inputs: T - positive number, temperature for Fermi Dirac distribution.
        Outputs: energy, nup, ndown

        Last Modified: 2020-09-16
        """
        # Generate the potential energy.
        _, _, offset = self._potential()
        # Solve for the single-electron energy levels.
        eup, edown, vup, vdown = self._eigensystem()
        # Get chemical potential
        energies = np.sort(np.concatenate((eup,edown)))
        mu = self._chemical_potential_from_states(
            T,(self.nelectup+self.nelectdown)*self.kpoints,energies)
        # Calculate energy
        en = self._energy_from_states(eup,edown,offset,T,mu)
        # Get new electron densities.
        weightup = fermi_distribution(eup,T,mu)
        weightdown = fermi_distribution(edown,T,mu)
        nup = np.sum(weightup*abs(vup)**2,1)/self.kpoints
        ndown = np.sum(weightdown*abs(vdown)**2,1)/self.kpoints
        # Return
        return en, nup, ndown
    #
    def _eigenstep_GCE(self,T,mu):
        """
        Evaluate the Hamiltonian, get new energy and density, in grand canonical ensemble

        Takes the electron densities and the parameters needed to
        make the Hamiltonian, then returns the energy and the
        electron densities predicted by the eigenvectors.
        Fills the eigenstates by assuming a Fermi-Dirac distribution
        with temperature T and chemical potential mu.

        Inputs: T - positive number, temperature for Fermi Dirac distribution.
            mu - number, chemical potential
        Outputs: energy, nup, ndown

        Last Modified: 2021-03-23
        """
        # Generate the potential energy.
        _, _, offset = self._potential()
        # Solve for the single-electron energy levels.
        eup, edown, vup, vdown = self._eigensystem()
        # Calculate energy
        en = self._energy_from_states(eup,edown,offset,T,mu)
        # Get new electron densities.
        weightup = fermi_distribution(eup,T,mu)
        weightdown = fermi_distribution(edown,T,mu)
        nup = np.sum(weightup*abs(vup)**2,1)/self.kpoints
        ndown = np.sum(weightdown*abs(vdown)**2,1)/self.kpoints
        # Apply GCE correction to energy (-mu*N)
        en -= mu * (nup.sum() + ndown.sum())
        # Return
        return en, nup, ndown
    #
    def _electron_density_single(self,n,method,up,alpha=None,**kwargs):
        """
        Helper method for set_electrons, so I don't have to
        duplicate my code for the spin up and down cases.

        Inputs: n - either a number of electrons
                or a list-like for the electron density.
                If number, must be between 0 and number of sites.
                If list-like, must be of right length and have
                values between 0 and 1.
            method - string. {'uniform', 'random'}. Specifies
                how to generate the electron density if a number
                of electrons is provided. Ignored if n is a list.
            up - Boolean, whether or not this is for the spin-up electrons.
                Doesn't do anything on its own, but may be passed to
                other methods.
        Keyword Arguments:
            alpha - optional positive number. Used for random method.
                Dirichlet alpha parameter. Default is chosen automatically.
        Outputs: density - ndarray of shape (3*nrows*ncols,)
            nelect - integer.
        Last Modified: 2021-02-19
        """
        # A useful constant
        nsites = self.nsites
        try:
            # First, we need to determine if n is the number of
            # electrons or a list representing the electron density.
            len(n)
        except TypeError:
            # n is the number of electrons.
            if not self.allow_fractions:
                nelect = int(round(n))
                # Check that n was at least close to an integer.
                if abs(nelect - n) > 1e-10:
                    warn("Number of electrons "+str(sum(n))+
                         " is not an integer. Rounding to "+str(nelect)+".")
            else:
                nelect = n
            if nelect < 0 or nelect > nsites:
                raise ValueError("The number of electrons is out of bounds.")
            # Generate an electron density by the appropriate method.
            density = self._electron_density_single_methods(nelect,method,up,
                                                            alpha=alpha,**kwargs)
        else:
            # n is the electron density.
            density = np.asarray(n, dtype=float)
            if density.min() < 0 or density.max() > 1:
                raise ValueError("The electron density is out of bounds.")
            if len(density) != nsites:
                raise ValueError("n has the wrong length.")
            if not self.allow_fractions:
                nelect = int(round(sum(n)))
                if abs(nelect - sum(n)) > 1e-10:
                    warn("Number of electrons "+str(sum(n))+
                         " is not an integer. Rounding to "+str(nelect)+".")
                    # Re-scale the density
                    sca = nelect/sum(n)
                    if sca < 1:
                        # Scale the electrons
                        density *= sca
                    else:
                        # Scale the holes
                        sca = (nsites - nelect)/(nsites - sum(n))
                        density = 1 - sca + sca*density
            else:
                nelect = sum(n)
        # Check that values are within bounds.
        # We had checks earlier checking the inputs, but here we double check.
        if nelect < 0 or nelect > nsites:
            raise ValueError("The number of electrons is out of bounds.")
        if density.min() < 0 or density.max() > 1:
            raise ValueError("The electron density is out of bounds.")
        return density, nelect
    #
    def _electron_density_single_methods(self,nelect,method,up,**kwargs):
        """
        Helper method for _electron_density_single.
        Handler for methods for specifying the electron density from
        the number of electrons.
        Should a child class wish to implement further methods, they should
        put `density=super()._electron_density_single_method(*args,**kwargs)'
        after checking for their own custom methods (in the else statement),
        to pass the execution back up the stack.
        
        Inputs: nelect - number of electrons. This has been tested for being
                a fraction or integer.
            method - a string. Specifies the method.
            up - boolean. Specifies whether this is a spin up or down electron.
        Keyword Arguments:
            alpha - positive number. Dirichlet alpha parameter for 'random' method.
            points - boolean. Parameter for 'impurity' method. Are up spins
                the localised ones?
        Output: electron density - (self.nsites,) ndarray, values between 0 and 1.
        Last Modified: 2021-03-02
        """
        # Process kwargs
        alpha = kwargs.get('alpha',None)
        points = kwargs.get('points',True)
        # Switch-case the method.
        if method == 'uniform':
            # Uniform electron density.
            density = nelect/self.nsites * np.ones(self.nsites)
        elif method == 'random':
            # Fully random electron density.
            density = random_density(self.nsites,nelect,alpha)
        elif method == 'impurity':
            # One spin channel is fully localised in random
            # sites while the other channel is uniform.
            if (points and not up) or (not points and up):
                # Uniform
                density = nelect/self.nsites * np.ones(self.nsites)
            else:
                # Random points
                density = random_points_density(self.nsites,nelect)
        elif method == 'scale':
            # Scale the existing electron density
            # Get the existing electron density.
            if up:
                nelect_old = self.nelectup
                density_old = self.nup
            else:
                nelect_old = self.nelectdown
                density_old = self.ndown
            # Scale the density.
            if nelect_old > nelect:
                # Reducing number of electrons, scale electrons
                density = density_old * nelect/nelect_old
            elif nelect_old < nelect:
                # Increasing number of electrons. Scale holes.
                # If scaled electrons, we'd risk going out of bounds.
                sca = (self.nsites - nelect)/(self.nsites - nelect_old)
                density = 1 - sca + sca * density_old
        else:
            # Cannot find the method. Raise error.
            raise ValueError("Method "+str(method)+" does not exist.")
        return density
    #
    ## GETTERS
    #
    def band_structure(self, klist, nsec):
        """
        Get the band structure - eigenenergies in paths between some k-points.

        Currently the path in the Brillouin zone needs to be continuous.
        I might eventually add support for discontinuous band structures.

        Inputs: klist - list-like containing length self.dim list-likes of
                numbers, i.e. k-points. k-points are in fractional coordinates.
                We compute the band structure in straight lines between
                consecutive points in klist.
            nsec - positive integer. Number of k-points to evaluate along each
                path (not including the starting point).
        Outputs: a shape (nsec*(len(klist)-1)+1,self.dim) ndarray, which is the
                list of all the k-points, in fractional coordinates.
            A shape (nsec*(len(klist)-1)+1,2,self.nsites) ndarray, which is the
                spin polarised band structure. 1st axis maps to the list of
                kpoints. 2nd axis is spin (up then down).

        Last Modified: 2020-10-02
        """
        # Convert to numpy array
        klist = np.asarray(klist)
        # Generate full list of k-points from the provided list of key points.
        # Piecewise parametric lines. The parameter within each segment,
        # (i/nsec - int(i/nsec)), has range [0,1).
        # It is essentially k[n+1]*t + k[n]*(1-t).
        klist_full = [ (klist[int(np.ceil(i/nsec))] - klist[int(i/nsec)])
                       * (i/nsec - int(i/nsec)) + klist[int(i/nsec)]
                       for i in range(nsec*(len(klist)-1)+1) ]
        # Return the band structure
        return klist_full, self.eigenvalues_at_kpoints(klist_full)
    #
    def band_structure_projected(self, klist, nsec, atoms):
        """
        Get the band structure with projections onto some atoms.

        Currently the path in the Brillouin zone needs to be continuous.
        I might eventually add support for discontinuous band structures.

        Inputs: klist - list-like containing length self.dim list-likes of
                numbers, i.e. k-points. k-points are in fractional coordinates.
                We compute the band structure in straight lines between
                consecutive points in klist.
            nsec - positive integer. Number of k-points to evaluate along each
                path (not including the starting point).
            atoms - list of integers, indices of sites/orbitals we want to see
                the projection onto.
        Outputs: klist_full - a shape (nsec*(len(klist)-1)+1,self.dim) ndarray,
                which is the list of all the k-points, in fractional coordinates.
            bands - A shape (nsec*(len(klist)-1)+1,2,self.nsites) ndarray, which
                is the spin polarised band structure. 1st axis maps to the list
                of kpoints. 2nd axis is spin (up then down).
            projections - ndarray, same shape as bands. The projections of each
                eigenstate in bands onto atoms.

        Last Modified: 2021-09-30
        """
        # Make the full list of k-points.
        klist = np.asarray(klist)
        klist_full = [ (klist[int(np.ceil(i/nsec))] - klist[int(i/nsec)])
                       * (i/nsec - int(i/nsec)) + klist[int(i/nsec)]
                       for i in range(nsec*(len(klist)-1)+1) ]
        bands, projections = self.eigenvalues_at_kpoints_projected(atoms, klist_full)
        return klist_full, bands, projections
    #
    def chemical_potential(self,T,N=None):
        """
        Determine chemical potential for a given temperature.
        
        Determines what the chemical potential should be
        to yield the desired number of electrons at a
        given temperature.

        Inputs: T - nonnegative number, temperature.
                If T==0, simply returns the energy of the
                highest occupied state.
            N - number of electrons. Defaults to however
                many the system currently has, times the number
                of k-points.
        Output: the chemical potential, a number
        Last Modified: 2020-09-16
        """
        if T < 0:
            raise ValueError("T cannot be negative.")
        if N is None:
            N = (self.nelectup + self.nelectdown)*self.kpoints
        # Get the eigenenergy spectrum.
        eup, edown, _, _ = self._eigensystem()
        # Consolidate and sort the eigenenergies
        energies = np.sort(np.concatenate((eup,edown)))
        return self._chemical_potential_from_states(T,N,energies)
    #
    def density_of_states(self, sigma, de, emin=None, emax=None, atoms=None):
        """
        Compute density of states with Gaussian smearing.

        Supports calculating projection onto atoms.

        Inputs: sigma - positive number, Gaussian smearing width.
            de - positive number, step-size for energy plotting.
            emin - number, optional. Minimum energy to plot.
                Defaults to minimum eigenvalue - 3 sigma.
            emax - number, optional. Maximum energy to plot.
                Defaults to maximum eigenvalue + 3 sigma.
            atoms - list of indices, optional. Atoms to project onto.
                Defaults to all atoms
        Outputs: Three numerical 1D ndarrays of equal length.
            energy - the energy axis.
            dosup - the DOS of the spin up states.
            dosdown - the DOS of the spin down states.

        Last Modified: 2022-11-03
        """
        # Get the eigenenergy spectrum.
        eup, edown, vup, vdown = self._eigensystem()
        # Get projection
        if atoms is None:
            pup = np.ones(eup.shape)
            pdown = np.ones(edown.shape)
        else:
            pup = (abs(vup[atoms,:])**2).sum(axis=0)
            pdown = (abs(vdown[atoms,:])**2).sum(axis=0)
        # Determine energy bounds if not specified.
        if emin is None:
            emin = min(eup.min(),edown.min()) - 3*sigma
        if emax is None:
            emax = max(eup.max(),edown.max()) + 3*sigma
        # Get the energy axis.
        energy = np.arange(emin,emax+1e-15,de) # +epsilon to include endpoint.
        # Gaussian smearing on energy levels gives DOS.
        # Get coordinate grids.
        energymesh, eupmesh = np.meshgrid(energy, eup, sparse=True)
        _, edownmesh = np.meshgrid(energy, edown, sparse=True)
        _, pupmesh = np.meshgrid(energy, pup, sparse=True)
        _, pdownmesh = np.meshgrid(energy, pdown, sparse=True)
        # Do Gaussian smearing.
        dosup = (pupmesh * np.exp(-(energymesh - eupmesh)**2 / (2*sigma))/
                 (sigma*np.sqrt(2*np.pi))).sum(axis = 0)
        dosdown = (pdownmesh * np.exp(-(energymesh - edownmesh)**2 / (2*sigma))/
                 (sigma*np.sqrt(2*np.pi))).sum(axis = 0)
        # Return
        return energy, dosup/self.kpoints, dosdown/self.kpoints
    #
    def eigenvalues_at_kpoints(self, klist=None):
        """
        Returns eigenenergies at the specified k-points.

        Input: klist, list-like containing length self.dim list-likes
                of numbers. i.e. a list of k-points. Defaults to self.kmesh.
                k-points are in fractional coordinates.
        Output: a shape (len(klist),2,self.nsites) ndarray. First axis is
            the k-points. 2nd axis is the spin (up or down). Third are the
            eigenergies.

        Last Modified: 2020-10-02
        """
        # Default value
        if klist is None:
            klist = self.kmesh
        # Initialise
        bands = np.empty((len(klist),2,self.nsites))
        potup, potdown, _ = self._potential()
        # Iterate through the k-points.
        for n, k in enumerate(klist):
            kin = self.get_kinetic(k)
            # Spin up eigenvalues
            bands[n,0] = np.linalg.eigvalsh(kin + potup)
            # Spin down eigenvalues
            bands[n,1] = np.linalg.eigvalsh(kin + potdown)
        return bands

    def eigenvalues_at_kpoints_projected(self, atoms, klist=None):
        """
        Returns eigenenergies at the specified k-points, projected onto atoms.

        Input: atoms - list of indices. Atoms to project onto.
            klist, list-like containing length self.dim list-likes
                of numbers. i.e. a list of k-points. Defaults to self.kmesh.
                k-points are in fractional coordinates.
        Output: bands - a shape (len(klist),2,self.nsites) ndarray. First axis is
                the k-points. 2nd axis is the spin (up or down). Third are the
                eigenergies.
            projections - ndarray of same shape as bands. Projections of each
                eigenstate onto the specified atoms.

        Last Modified: 2021-09-30
        """
        if klist is None:
            klist = self.kmesh
        bands = np.empty((len(klist), 2, self.nsites))
        projections = np.empty(bands.shape)
        # Got to do eigensysteming manually
        potup, potdown, _ = self._potential()
        for n, k in enumerate(klist):
            kin = self.get_kinetic(k)
            # Spin up
            bands[n,0], v = np.linalg.eigh(kin + potup)
            projections[n,0] = (abs(v[atoms,:])**2).sum(axis=0)
            # Spin down
            bands[n,1], v = np.linalg.eigh(kin + potdown)
            projections[n,1] = (abs(v[atoms,:])**2).sum(axis=0)
        return bands, projections
    #
    def eigenstates(self,mode='states',emin=None,emax=None):
        """
        Returns a selection of band-decomposed electron densities.
        
        Defaults to returning all of them.
        Currently lacks the means to distinguish between states
        from different k-points.

        Inputs: mode - string. How to interpret emin and emax.
                'states': as 1-based indices.
                'energy': as energies.
                'fermi': as energies relative to the Fermi level.
            emin - number. Minimum band to plot.
            emax - number. Maximum band to plot.
        Outputs:
            List of eigenenergies of spin up states.
            List of electron densities of spin up states.
            List of eigenenergies of spin down states.
            List of electron densities of spin down states.

        Last Modified: 2020-09-16
        """
        # Get eigenstates
        eup, edown, vup, vdown = self._eigensystem()
        # Convert emin and emax to indices
        if mode=='states': # emin and emax are indices, 1-based indexing.
            if emin is None:
                emin = 1
            if emax is None:
                emax = 3*self.nrows*self.ncols
            iminup = int(emin-1)
            imaxup = int(emax-1)
            imindown = int(emin-1)
            imaxdown = int(emax-1)
        elif mode=='energy': # emin and emax are energies.
            if emin is None:
                emin = min(eup.min(),edown.min())
            if emax is None:
                emax = max(eup.max(),edown.max())
            iminup = (eup >= emin).nonzero()[0][0]
            imaxup = (eup <= emax).nonzero()[0][-1]
            imindown = (edown >= emin).nonzero()[0][0]
            imaxdown = (edown <= emax).nonzero()[0][-1]
        elif mode=='fermi': # they are energies relative to the Fermi energy.
            fermiup, fermidown = self.fermi()
            if emin is None:
                emin = min(eup.min()-fermiup,edown.min()-fermidown)
            if emax is None:
                emax = max(eup.max()-fermiup,edown.max()-fermidown)
            iminup = (eup >= (emin+fermiup)).nonzero()[0][0]
            imaxup = (eup <= (emax+fermiup)).nonzero()[0][-1]
            imindown = (edown >= (emin+fermidown)).nonzero()[0][0]
            imaxdown = (edown <= (emax+fermidown)).nonzero()[0][-1]
        else:
            raise ValueError('Mode '+str(mode)+' is not recognised.')
        return (eup[iminup:imaxup+1],
                abs(vup.transpose()[iminup:imaxup+1])**2,
                edown[imindown:imaxdown+1],
                abs(vdown.transpose()[imindown:imaxdown+1])**2)
    #
    def eigensystem(self):
        """
        Solve the eigensystem and return the eigenenergies and states.

        Outputs: eup, edown, vup, vdown
            eup, edown - (nsites*kpoints,) ndarray of numbers, sorted.
                    Eigenenergies.
            vup, vdown - (nsites,nsites*kpoints) ndarray of numbers.
                Eigenvectors.
                v[:,i] is the normalised eigenvector corresponding to the
                eigenvalue e[i], as output by np.linalg.eigh.

        Last Modified: 2020-09-16
        """
        # Grab constants for conciseness
        nsites = self.nsites
        # Get the potential matrices
        potup, potdown, _ = self._potential()
        # Initialise
        eup = np.empty(nsites * self.kpoints)
        edown = np.empty(nsites * self.kpoints)
        vup = np.empty((nsites, nsites * self.kpoints),dtype='complex')
        vdown = np.empty((nsites, nsites * self.kpoints),dtype='complex')
        # Solve for the single-electron energy levels at each k-point
        for n, k in enumerate(self.kmesh):
            kin = self.get_kinetic(k)
            eup[n*nsites:(n+1)*nsites], vup[:,n*nsites:(n+1)*nsites] =\
                                        np.linalg.eigh(kin + potup)
            edown[n*nsites:(n+1)*nsites], vdown[:,n*nsites:(n+1)*nsites] =\
                                        np.linalg.eigh(kin + potdown)
        # Sort eigenstates by energy.
        # Skip this step if only one k-point due to redundancy.
        if self.kpoints > 1:
            inds_up = np.argsort(eup)
            inds_down = np.argsort(edown)
            eup = eup[inds_up]
            edown = edown[inds_down]
            vup = vup[:,inds_up]
            vdown = vdown[:,inds_down]
        return eup, edown, vup, vdown
    # Alias for backwards compatability
    _eigensystem = eigensystem
    #
    def energy(self,T=None, mu=None):
        """
        Calculates the energy from the current electron density.
        Does not go through self-consistency.

        Input: T - optional positive number. If provided,
            energy is calculated with a Fermi distribution
            for the occupancies.
            mu - optional number. If provided, T must be provided.
            Chemical potential. Calculates energy within grand canonical
            ensemble.
        Output: energy, a number.

        Last Modified: 2021-03-23
        """
        # Generate the potential energy.
        _, _, offset = self._potential()
        # Extra offset if grand canonical ensemble
        if mu is not None:
            offset -= mu * self.get_electron_number()
        # Solve for the single-electron energy levels.
        eup, edown, _, _ = self._eigensystem()
        if T is not None:
            if mu is None:
                mu = self.chemical_potential(T)
        else:
            if mu is not None:
                raise TypeError("If mu is provided, T must be provided.")
        return self._energy_from_states(eup,edown,offset,T=T,mu=mu)
    #
    def fermi(self, midpoint=False):
        """
        Calculates the Fermi level for the single particle states.
        
        Really, it just returns the energy of the highest occupied states.
        If midpoint=True, it returns the energy in between the highest occupied
        and lowest unoccupied states.
        It's probably a better idea to use chemical_potential for most uses.

        Outputs: two numbers, the Fermi level for spin up and spin down states.

        Last Modified: 2021-02-17
        """
        # Solve for the single-electron energy levels.
        eup, edown, _, _ = self._eigensystem()
        if self.allow_fractions:
            Nup = int(round(self.nelectup*self.kpoints))
            Ndown = int(round(self.nelectdown*self.kpoints))
        else:
            Nup = self.nelectup*self.kpoints
            Ndown = self.nelectdown*self.kpoints
        # Check the highest occupied states for the Fermi.
        # Get spin up
        if Nup == 0:
            fermi_up = eup[0]
        elif midpoint and len(eup) < Nup:
            fermi_up = (eup[Nup] + eup[Nup-1])/2
        else:
            fermi_up = eup[Nup-1]
        # get spin down
        if Ndown == 0:
            fermi_down = edown[0]
        elif midpoint and len(edown) < Ndown:
            fermi_down = (edown[Ndown] + edown[Ndown-1])/2
        else:
            fermi_down = edown[Ndown-1]
        # Return fermi
        return fermi_up, fermi_down
    #
    def get_charge_density(self):
        """
        Returns the charge density.
        Output: (3*nrows*ncols,) ndarray.
        Last Modified: 2020-07-13
        """
        return self.nup + self.ndown
    #
    def get_coordinates(self):
        """
        MUST BE OVERRIDDEN IN SUBCLASS.
        Returns the coordinates for plotting electron density.
        Output: a (self.nsites,2) ndarray of floats.
        """
        raise NotImplementedError
    #
    def get_electron_number(self):
        """
        Return the number of electrons.
        
        Output: Number
        Last Modified: 2020-07-29
        """
        return self.nelectup + self.nelectdown
    #
    def get_kinetic(self,k):
        """
        MUST BE OVERRIDDEN IN SUBCLASS.
        
        Returns the kinetic energy matrix for a given momentum
        Expected Input: k, a length self.dims vector.
        Expected Output: (nsites,nsites) ndarray
        Last Modified: 2020-09-16
        """
        raise NotImplementedError
    #
    def get_magnetization(self):
        """
        Return the net spin.
        
        Output: Number, the net spin.
        Last Modified: 2020-07-20
        """
        return self.nelectup - self.nelectdown
    #
    def get_spin_density(self):
        """
        Return the spin density.
        
        Output: (3*nrows*ncols,) ndarray.
        Last Modified: 2020-07-13
        """
        return self.nup - self.ndown
    #
    def local_magnetic_moment(self):
        """
        Return the average value of the local magnetic moment squared.
        
        Output: Number
        Last Modified: 2020-07-03
        """
        return (self.get_spin_density()**2).mean()
    #
    def nelect_from_chemical_potential(self, mu, T):
        """
        Solves for number of electrons to get a chemical potential mu.

        Not self-consistent. You've got to figure that one out yourself.

        Inputs: mu - number, chemical potential
            T - non-negative number, temperature for Fermi-Dirac distribution
        Output: N - non-negative number, number of electrons.

        Last Modified: 2021-03-02
        """
        # Get the eigenstates
        eup, edown, _, _ = self._eigensystem()
        # Concatenate
        energies = np.sort(np.hstack((eup, edown)))
        # Solve
        N = fermi_distribution(energies,T,mu).sum()/self.kpoints
        # Put within valid range
        N = min(max(N, 0), 2*self.nsites)
        return N
    #
    def residual(self,T=None,mu=None):
        """
        Return the residual of the electron density.
        
        A key convergence figure.
        Self-consistency has been reached if this number
        is 'small', ideally 0.

        Input: T - optional number. Temperature for eigenstep.
            mu - optional number. Chemical potential for GCE eigenstep.
        Outputs: residual , a non-negative number

        Last Modified: 2021-03-23
        """
        # Get densities after one iteration.
        if T is None:
            _, nupnew, ndownnew = self._eigenstep()
        else:
            if mu is None:
                _, nupnew, ndownnew = self._eigenstep_finite_T(T)
            else:
                _, nupnew, ndownnew = self._eigenstep_GCE(T,mu)
        # Calculate residual squared.
        residual_up = nupnew - self.nup
        residual_down = ndownnew - self.ndown
        res = np.linalg.norm(np.concatenate((residual_up,residual_down)))
        return res
    #
    def _chemical_potential_from_states(self,T,N,energies):
        """
        Determine chemical potential for a given temperature and set of states.

        I note that my T=0 algorithm sometimes gives different results to the
        limit T->0.
        
        Inputs: T - nonnegative number, temperature.
                If T==0, simply returns the energy of the
                highest occupied state.
            N - number of occupied states (electrons * kpoints).
            energies - sorted list/ndarray of all eigenenergies.
        Output: the chemical potential, a number
        Last Modified: 2021-02-18
        """
        if T < 0:
            raise ValueError("T cannot be negative.")
        # Check impossible cases
        if N > len(energies) or N < 0:
            raise ValueError("N is out of bounds.")
        # Special case: T=0
        if T==0:
            # At zero temperature, we cannot have fractional
            # occupation. N must be an integer.
            N = int(round(N))
            # Get the N'th eigenenergy.
            if N == 0:
                mu = energies[0]
            else:
                mu = energies[N-1]
        else:
            # Convert to array so it works in fermi_distribution
            energies = np.asarray(energies)
            # Get a first guess of the chemical potential.
            mu = energies[int(round(N))-1]
            nguess = fermi_distribution(energies,T,mu).sum()
            # Get a good initial step-size
            # Should either be comparable to the energy gap or
            # the temperature.
            ediff = np.diff(energies)
            i = int(round(N)) - 1 # index
            if i >= len(ediff):
                step = max(T,ediff[-1])
            elif i <= 0:
                step = max(T,ediff[0])
            else:
                step = max(T,ediff[i],ediff[i-1])
            # Perform a binary search to find mu.
            first_sweep = True
            start_low = (N < nguess)
            while abs(N-nguess) > 1e-12 and step > np.spacing(mu):
                # Stop either when we have equality or our step size has
                # shrunk into oblivion and we cannot get any closer at
                # machine precision.
                # Check if we have found a cross-over point.
                if first_sweep:
                    if ((start_low and N > nguess) or
                        (not start_low and N < nguess)):
                        first_sweep = False
                if not first_sweep:
                    step /= 2 # Halve step-size for binary search.
                # Increment chemical potential in correct direction.
                if N > nguess:
                    mu += step
                else:
                    mu -= step
                # Recalculate our guess at the number of electrons.
                nguess = fermi_distribution(energies,T,mu).sum()
        return mu
    #
    def _energy_from_states(self,eup,edown,offset,T=None,mu=None):
        """
        Calculate energy from pre-calculated eigenstates.
        
        Inputs:
            eup - ndarray of spin up eigenstates.
            edown - ndarray of spin down eigenstates
            offset - number, energy offset from _potential
            T - number, optional. Temperature for Fermi distribution
                If not provided, just fill lowest energy states
                for each spin sector.
            mu - number, optional. Chemical potential. Must be provided
                if T is provided.
        Last Modified: 2020-09-16
        """
        if T is None:
            # Number of states.
            Nup = self.nelectup * self.kpoints
            Ndown = self.nelectdown * self.kpoints
            # Calculate the energy by filling electrons.
            if not self.allow_fractions:
                # Occupation is strictly integers.
                en = (sum(eup[0:Nup]) + sum(edown[0:Ndown])) / self.kpoints \
                     + offset
            else:
                # Occupation may be fractional.
                en = sum(eup[0:int(Nup)]) + sum(edown[0:int(Ndown)])
                # Get the fractional part of the occupations.
                fracup = Nup - int(Nup)
                fracdown = Ndown - int(Ndown)
                # Add energy from fractionally occupied states.
                if fracup > 0: en += eup[int(Nup)] * fracup
                if fracdown > 0: en += edown[int(Ndown)] * fracdown
                en = en / self.kpoints + offset
        else:
            # Calculate energies by weighting states by Fermi distribution
            if mu is None:
                raise TypeError("If T is provided, mu must also be provided.")
            energies = np.concatenate((eup,edown))
            en = np.sum(energies *
                        fermi_distribution(energies, T, mu))/self.kpoints \
                        + offset
        return en
    #
    def _potential(self):
        """
        Create the mean-field Hubbard potential part of the Hamiltonian.
        
        Give it the densities/occupations of up and down electrons,
        and it will give the Hubbard potential for up and down electrons.

        Presets: u, nup, ndown, mag
        Outputs: potup, potdown, offset
            two square diagonal ndarrays, of dimension equal to nup, and a number.
            potup is potential for spin up electrons.
            potdown is potential for spin down electrons.
            offset is a real number, to be added to the total energy.

        Last Modified: 2021-02-17
        """
        # V = U(nup<ndown> + ndown<nup> - <nup><ndown>)
        # Also, the magnetic field shifts the energy of up and down electrons.
        # We have <nup> and <ndown>
        offset = -np.sum(self.u*self.nup*self.ndown) # -u <nup><ndown>
        potup = self.u*np.diag(self.ndown) - self.mag * np.eye(self.nsites)
        potdown = self.u*np.diag(self.nup) + self.mag * np.eye(self.nsites)
        return potup, potdown, offset
    #
    ## PLOTTERS
    #
    def plot_bands(self, klist, nsec, T=None, emin=None, emax=None,
                   klabels=None, atoms=None, inplace=True, ax=None):
        """
        Plot the band structure, for some given k-points.

        Inputs: klist - list-like containing length self.dim list-likes of
                numbers, i.e. k-points. k-points are in fractional coordinates.
                We compute the band structure in straight lines between
                consecutive points in klist.
            nsec - positive integer. Number of k-points to evaluate along each
                path (not including the starting point).
            T - non-negative number. Temperature with which to calculate
                the chemical potential. If omitted, no Fermi energy is used.
                If provided, energies are shifted to be relative to Fermi.
            emin - number. Minimum energy to plot.
            emax - number. Maximum energy to plot.
            klabels - length len(klist) list of strings, marking the tick
                labels.
            atoms - optional. List of integers. Indices of sites/orbitals to
                project onto. Shows as circles on the band structure.
            inplace - Boolean. Plot in place.
            ax - optional, matplotlib Axes. Plot to this.
        Effects: Makes a plot.
        Returns: Figure, Axes.

        Last Modified: 2021-09-13
        """
        # Get the band structure.
        if atoms is None:
            kp, bands = self.band_structure(klist, nsec)
        else:
            kp, bands, projections = self.band_structure_projected(klist, nsec, atoms)
        # Shift the bands by the chemical potential
        if T is not None:
            bands -= self.chemical_potential(T)
        # Convert the k-points from fractional to cartesian coords,
        # then turn it into a 1D distance.
        kp = np.matmul(kp, self.reclat)
        klen = np.cumsum(np.linalg.norm(np.diff(kp, axis=0, prepend=kp[0:1]),
                                        axis=1))
        # Identify the tic locations
        tics = klen[np.arange(0, nsec*(len(klist)-1)+1, nsec)]
        # Get the Axes and Figure.
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        # Plot spin up bands
        for b in bands[:,0,:].transpose():
            ax.plot(klen,b,color='r')
        # Plot spin down bands
        for b in bands[:,1,:].transpose():
            ax.plot(klen,b,color='b')
        # Plot projection onto atoms.
        if atoms is not None:
            for (b,p) in zip(bands[:,0,:].transpose(), projections[:,0,:].transpose()):
                ax.scatter(klen,b,s=p*36, edgecolor='r', facecolor='none', zorder=2.1, marker='o')
            for (b,p) in zip(bands[:,1,:].transpose(), projections[:,1,:].transpose()):
                ax.scatter(klen,b,s=p*36, edgecolor='b', facecolor='none', zorder=2.1, marker='o')
        ax.set_ylim(emin, emax)
        ax.set_xticks(tics)
        ax.set_xticklabels(klabels)
        ax.grid(axis='x')
        ax.axhline(y=0, color='0.5', linewidth=0.5, zorder=1.5)
        if inplace:
            plt.show()
        return fig, ax
    #
    def plot_spin(self, marker_scale=1, inplace=True, ax=None):
        """
        Plots the spin density.
        Size/area of marker is proportional to magnitude.
        Yellow is spin up. Blue is spin down.

        Inputs: marker_scale - number, optional. Factor to scale markersize by.
            inplace - Boolean. Plot in place.
            ax - optional, matplotlib Axes. Plot to this.
        Effects: Makes a plot.
        Output: Figure, Axes.

        Last Modified: 2021-09-13
        """
        # Get the Cartesian coordinates of each point.
        coords = self.get_coordinates()
        x = coords[:,0]
        y = coords[:,1]
        # Get the spin moment.
        spin = (self.nup - self.ndown)[:len(coords)]
        # Convert sign of spin to colour.
        color_up = 'y'
        color_down = 'b'
        spin_col = [ color_down if x<0 else color_up for x in np.sign(spin) ]
        # Get figure and axes
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        # Default marker size is 6. Default marker area is 6**2=36.
        ax.scatter(x,y,abs(spin)*36*(marker_scale**2),spin_col)
        ax.set_aspect('equal')
        if inplace:
            plt.show()
        return fig, ax
    #
    def plot_charge(self,marker_scale=1, inplace=True, ax=None):
        """
        Plots the charge density.
        Marker size/area is proportional to the charge density.

        Inputs: marker_scale - number, optional. Factor to scale markersize by.
            inplace - Boolean. Plot in place.
            ax - optional, matplotlib Axes. Plot to this.
        Effect: Makes a plot.
        Output: Figure, Axes

        Last Modified: 2021-09-13
        """
        # Get the Cartesian coordinates of each point.
        coords = self.get_coordinates()
        x = coords[:,0]
        y = coords[:,1]
        # Get the charge density.
        chg = (self.nup + self.ndown)[:len(coords)]
        # Get figure and axes
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        # Default marker size is 6. Default marker area is 6**2=36.
        ax.scatter(x,y,chg*36*(marker_scale**2))
        ax.set_aspect('equal')
        if inplace:
            plt.show()
        return fig, ax
    #
    def plot_DOS(self, sigma, de, emin=None, emax=None, midpoint=True,
                inplace=True, ax=None):
        """
        Plots the density of states. See density_of_states for arguments.
        Also shows the Fermi levels, although this Fermi level does
        not fully account for the broadening.
        (I wrote this Fermi calculation prior to the fancier calculations
        I use later. As such, this function is slightly antiquated.)

        Inputs:
            sigma - positive number. Gaussian broadening to apply.
            de - positive number. Energy step to sample.
            emin - optional number. Minimum energy bound (defaults to full range)
            emax - optional number. Maximum energy bound (defaults to full range)
            midpoint - Boolean. Determines whether the Fermi level is given by
                the midpoint between the highest occupied and lowest unoccupied
                states, or as just the highest occupied state.
            inplace - Boolean. Whether to plot in-place.
            ax - optional matplotlib Axes to plot to.
        Output: Figure, Axes.

        Last Modified: 2021-09-13
        """
        # Calculate the DOS.
        energy, dosup, dosdown = self.density_of_states(sigma,de,emin,emax)
        # Calculate the Fermi level.
        fermi_up, fermi_down = self.fermi(midpoint)
        # Get figure and axes
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.figure
        # Plot the DOS.
        ax.plot(energy, dosup, 'k', energy, -dosdown, 'k') # Black lines
        ax.set_xlabel('Energy')
        ax.set_ylabel('DOS')
        # Plot the Fermi
        ax.vlines(fermi_up,0,dosup.max(),colors='b')
        ax.vlines(fermi_down,-dosdown.max(),0,colors='b')
        # Show the plot.
        if inplace:
            plt.show()
        return fig, ax
    #
    def plot_spincharge(self,marker_scale=1,scale=1,cmap='BrBG_r',
                    inplace=True, ax=None, cbar=True, swap_xy=False):
        """
        Plots the spin and charge density.
        Marker size/area is proportional to the charge density.
        Marker colour is proportional to the spin density (yellow up, blue down).

        Inputs: marker_scale - number, optional. Factor to scale markersize by.
            scale - number (0,1], optional. Spin density is plotted from 
                -scale to scale.
            cmap - str or matplotlib Colormap. Colormap name or instance.
                I recommend a divergent colormap.
            inplace - Boolean. Whether to plot in-place.
            ax - optional matplotlib Axes to plot to.
            cbar - Boolean. Plot the colourbar.
            swap_xy - Boolean. Swap x and y coords.
        Effect: Makes a plot
        Output: Figure, Axes

        Last Modified: 2021-09-15
        """
        # Get the Cartesian coordinates of each point.
        coords = self.get_coordinates()
        if swap_xy:
            y,x = coords[:,0], coords[:,1]
        else:
            x,y = coords[:,0], coords[:,1]
        # Get the charge density.
        chg = (self.nup + self.ndown)[:len(coords)]
        # Get the spin moment.
        spin = (self.nup - self.ndown)[:len(coords)]
        # Get figure and axes
        if ax is None:
            # Set up figure size to reduce whitespace
            # What aspect ratio do we need?
            if y.max() == y.min():
                aspect = 1000
            else:
                aspect = (x.max() - x.min())/(y.max() - y.min())
            # Bound the aspect ratio to prevent pathological cases.
            aspect = max(min(aspect, 4), 1/4)
            # Go from default figsize...
            figsize = plt.rcParams["figure.figsize"]
            default_aspect = figsize[0]/figsize[1]
            if aspect > default_aspect:
                # Figure is short and wide
                figsize = [figsize[0], figsize[0]/aspect]
            else:
                # Figure is skinny and tall
                figsize = [figsize[1]*aspect, figsize[1]]
            if cbar:
                # Add some space for the colourbar.
                figsize[0] += 0.5
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.figure
        # Plot
        scat = ax.scatter(x,y,chg*36*(marker_scale**2),spin,cmap=cmap,vmin=-scale,vmax=scale)
        ax.set_aspect('equal')
        # Remove ticks
        ax.set_xticks([])
        ax.set_yticks([])
        # Draw colourbar.
        if cbar:
            cb = fig.colorbar(scat, ax=ax)
            cb.ax.set_title('Spin')
        if inplace:
            plt.show()
        return fig, ax
    #
    ## SETTERS
    #
    def set_kinetic(self):
        """
        Use this method to set self.kin (for Gamma-point only) or
        the parameters to generate the kinetic energy from.
        Implementation is lattice-dependent.
        """
        raise NotImplementedError
    #
    def set_kmesh(self,*args,method='monkhorst'):
        """
        Set the k-mesh used for Brillouin zone sampling.

        Uses fractional coordinates.
        There are two ways to pass arguments.
        The first way is to specify self.dims integers, which produces a
        Monkhorst pack grid.
        The second way is to provide an explicit list of k-points.

        Inputs (1): self.dims positive integers.
        Inputs (2): list of lists of length self.dims numbers.

        Keyword arguments: method - string, {'monkhorst', 'gamma'}
                'monkhorst' makes a regular Monkhorst Pack grid.
                'gamma' makes a Gamma-centred Monkhorst Pack grid.

        Sets self.kmesh (array of k-points, shape (self.kpoints,self.dims))
         and self.kpoints (number of k-points).

        Last Modified: 2020-09-16
        """
        # Identify the nature of our input arguments.
        # If argument is integer, make Monkhorst Pack grid.
        if isinstance(args[0],int):
            if len(args) != self.dims:
                raise TypeError("set_kmesh expected "+str(self.dims)+
                                " integer arguments, got "+str(len(args)))
            # Check that args are all positive integers.
            for i in args:
                if not isinstance(i,int) or i <= 0:
                    raise ValueError("Expected arguments"
                                     " to be positive integers.")
            # Create a list of evenly spaced k-points.
            coords = [np.arange(0,1,1/n) for n in args]
            self.kpoints = np.prod(args)
            self.kmesh = np.array(np.meshgrid(*coords))\
                         .reshape((2,self.kpoints)).transpose()
            if method == 'monkhorst':
                # In original Monkhorst-Pack scheme, an even number
                # of subdivisions leads to shifting away from k-point
                # by half a subdivision
                shift = np.mod(np.array(args)+1,2)/2/np.array(args)
                self.kmesh += shift
            elif method == 'gamma':
                pass
            else:
                raise TypeError("set_kmesh doesn't recognise method "
                                +str(method))
        else:
            # Assume argument to be an explicit list of k-points.
            self.kmesh = args[0]
            self.kpoints = len(self.kmesh)
    #
    def set_mag(self,mag):
        """
        Sets the magnetic field. +ve favours spin up.
        Input: mag - real number
        Effect: sets self.mag to mag
        Last Modified: 2020-07-29
        """
        self.mag = mag
    #
    def set_u(self,u):
        """
        Sets the Hubbard U parameter.

        Input: u - real number.
        Effect: Sets self.u to u.

        Last Modified: 2020-07-10
        """
        self.u = u
    #
    def toggle_allow_fractions(self,val=None):
        """
        Toggles flag which allows fractional electrons, handling conversion.
        
        Handles converting the number of electrons back to an integer if
        necessary.

        Inputs: val - optional Boolean. Sets allow_fractions
            to this value if provided. Toggles allow_fractions otherwise.
        Output: Boolean, new value of allow_fractions.

        Last Modified: 2021-02-18
        """
        if val is None:
            # If val not set, we toggle.
            val = not self.allow_fractions
        if val:
            # Setting to True is easy. Nothing to be done.
            self.allow_fractions = True
        else:
            # Setting to False is harder. Got to make sure
            # the electrons are all proper integers.
            nelectup_old = self.nelectup
            nelectdown_old = self.nelectdown
            # Set the number of electrons to integers.
            self.nelectup = int(round(self.nelectup))
            self.nelectdown = int(round(self.nelectdown))
            # Rescale the electron densities to match.
            if nelectup_old > 0:
                sca = self.nelectup/nelectup_old
                if sca < 1:
                    self.nup *= sca
                elif sca > 1:
                    # Do scaling of hole density instead.
                    sca = (self.nsites - self.nelectup)/(self.nsites - nelectup_old)
                    self.nup = 1 - sca + sca * self.nup
            if nelectdown_old > 0:
                sca = self.nelectdown/nelectdown_old
                if sca < 1:
                    self.ndown *= sca
                elif sca > 1:
                    # Do scaling of hole density instead.
                    sca = (self.nsites - self.nelectdown)/(self.nsites - nelectdown_old)
                    self.ndown = 1 - sca + sca * self.ndown
            # Now toggle the flag.
            self.allow_fractions = False
        return self.allow_fractions




#!/usr/bin/python3

"""
Mean field Hubbard model of the Kagome lattice, in a single unit cell.

Has periodic boundary conditions and explores full Brillouin zone.

Created: 2020-09-17
Last Modified: 2021-05-18
Author: Bernard Field
    Copyright (C) 2021 Bernard Field, GNU GPL v3+
"""

import json
from math import cos, pi

import numpy as np

from hubbard.kagome import kagome_coordinates, kagome_adjacency_tensor
from hubbard.kpoints.base import HubbardKPoints


class KagomeHubbardKPoints(HubbardKPoints):
    """Mean field Hubbard model of the Kagome lattice, in a single unit cell."""
    #
    ## IO
    #
    def __init__(self,u=0,t=1,nup=0,ndown=0,allow_fractions=False,
                 nrows=1,ncols=1,**kwargs):
        """
        Create a kagome lattice and initialises things.

        Last Modified: 2020-11-06
        """
        self.nrows = nrows
        self.ncols = ncols
        self.reclat = np.array([[np.sqrt(3)/2, 1/2], [-np.sqrt(3)/2, 1/2]])
        super().__init__(3*nrows*ncols, u=u, nup=nup, ndown=ndown,
                         allow_fractions=allow_fractions, **kwargs)
        self.set_kinetic(t)
    #
    def copy(self):
        """
        Return a copy of this kagome lattice object.

        Output: KagomeHubbardKPoints object.

        Last Modified: 2020-11-13
        """
        kagome = KagomeHubbardKPoints(u=self.u, t=self.t, nup=self.nup.copy(),
                                      ndown=self.ndown.copy(), nrows=self.nrows,
                                      ncols=self.ncols,
                                      allow_fractions=self.allow_fractions)
        kagome.set_kmesh(self.kmesh)
        kagome.set_mag(self.mag)
        if self.modified_kin:
            kagome.set_kinetic(self.t, force_generic=True)
            kagome.kin = self.kin.copy()
        return kagome
    #
    @classmethod
    def load(cls,f):
        """
        Load a KagomeHubbardKPoints object from a JSON file f.

        Inputs: f - string, filename.
        Output: KagomeHubbardKPoints object.
        Last Modified: 2020-11-11
        """
        # Load the file
        with open(f) as file:
            di = json.load(file) # di for dictionary
        # Create a KagomeHubbardKPoints object
        kagome = cls(u=di['u'], t=di['t'], nrows=di.get('nrows',1),
                     ncols=di.get('ncols',1), nup=np.asarray(di['nup']),
                     ndown=np.asarray(di['ndown']),
                     allow_fractions=di['allow_fractions'])
        kagome.set_kmesh(np.asarray(di['kmesh']))
        kagome.set_mag(di['mag'])
        if 'kin' in di:
            kagome.set_kinetic(di['t'], force_generic=True)
            kagome.kin = np.asarray(di['kin'], dtype='complex')
            kagome.modified_kin = True
        return kagome
    #
    def save(self,f):
        """
        Save a JSON representation of the object's data to file f.

        Inputs: f - string, filename.
        Writes a test file f.
        Last Modified: 2021-05-18
        """
        di = {
                'u' : float(self.u),
                't' : float(self.t),
                'nrows' : int(self.nrows),
                'ncols' : int(self.ncols),
                'mag' : float(self.mag),
                'nup' : self.nup.tolist(),
                'ndown' : self.ndown.tolist(),
                'allow_fractions' : bool(self.allow_fractions),
                'kmesh' : self.kmesh.tolist(),
                }
        if self.modified_kin:
            di['kin'] = self.kin.real.tolist()
        with open(f,mode='w') as file:
            json.dump(di, file)
    #
    ## ELECTRON DENSITY MODIFIERS
    #
    #
    ## GETTERS
    #
    def get_coordinates(self):
        """
        Return the coordinates for plotting electron density.

        Output: a (self.nsites,2) ndarray of floats.
        Last Modified: 2020-09-17
        """
        return kagome_coordinates(self.nrows,self.ncols)
    #
    def get_kinetic(self,k):
        """
        Return the kinetic energy matrix for a given momentum.

        For the single-unit-cell case,
        I use the real form of the matrix to avoid unnecessary complex numbers.
        This results in a gauge transformation on the eigenvectors, which
        only matters if you care about the phase (which I don't).
        
        For the supercell, I need to use the complex formulation.
        
        Input: k, a length 2 list-like of numbers, representing the momentum
                in fractional coordinates.
        Output: a (nsites,nsites) symmetric real ndarray.
        Last Modified: 2020-11-11
        """
        if self.nsites == 3 and not self.modified_kin:
            # The simple case of a single unit cell.
            c0 = -2 * self.t * cos(pi*k[0])
            c1 = -2 * self.t * cos(pi*k[1])
            c2 = -2 * self.t * cos(pi*(k[0]-k[1]))
            return np.array([[0,c0,c1],[c0,0,c2],[c1,c2,0]])
        else:
            # Multiple unit cells. Need to include phase factors at boundaries.
            kinetic = self.kin.copy()
            # e^ikb1
            kinetic[self.masks[0]] *= np.exp(complex(0,2*np.pi*k[0]))
            # e^ikb2
            kinetic[self.masks[1]] *= np.exp(complex(0,2*np.pi*k[1]))
            # e^ik(b2-b1)
            kinetic[self.masks[2]] *= np.exp(complex(0,2*np.pi*(k[1]-k[0])))
            # e^-ik(b2-b1)
            kinetic[self.masks[3]] *= np.exp(complex(0,-2*np.pi*(k[1]-k[0])))
            # e^-ikb2
            kinetic[self.masks[4]] *= np.exp(complex(0,-2*np.pi*k[1]))
            # e^-ikb1
            kinetic[self.masks[5]] *= np.exp(complex(0,-2*np.pi*k[0]))
            # Add back in the previously removed double-counted part.
            kinetic[self.two_mask] += self.kin[self.two_mask]
            return kinetic
    #
    ## PLOTTERS
    #
    #
    ## SETTERS
    #
    def set_kinetic(self, t, force_generic=False):
        """
        Set the hopping constant for the Hamiltonian.
        
        If nrows or ncols > 1, also pre-calculates the connectivity.

        Inputs: t - real number. Hopping constant.
        Effect: sets self.t to t
        Last Modified: 2020-11-11
        """
        self.t = t
        self.modified_kin = force_generic
        # Pre-compute as much as possible if we have a non-simple case.
        if self.nsites > 3 or force_generic:
            kinshape = (self.nsites,self.nsites)
            # Get the Gamma-point matrix.
            tensor = kagome_adjacency_tensor(self.nrows,self.ncols)
            # The case when nrows or ncols == 1 has some extra problems,
            # because you will get two different processes (one within the
            # supercell, and one from without) occupying the same matrix
            # element. These have a value of 2 in tensor.
            # We shall separate them out so we can treat them separately.
            self.two_mask = np.reshape((tensor == 2),kinshape,order='F')
            tensor[tensor==2] = 1
            self.kin = np.reshape(-t*tensor,kinshape,order='F').astype('complex')
            # Also have to cast to complex dtype, to allow complex numbers.
            # The k-dependence can be done as an elementwise multiplication.
            # Find where we cross over the boundaries.
            self.masks = []
            nrows = self.nrows
            ncols = self.ncols
            idrow = np.identity(nrows,dtype='bool').reshape(1,nrows,1,1,nrows,1)
            idcol = np.identity(ncols,dtype='bool').reshape(1,1,ncols,1,1,ncols)
            # (0,i,j)->(1,i-1,j), (2,i,j)->(1,i-1,j+1), needs e^ikb1
            m = np.zeros((3,nrows,1,3,nrows,1), dtype='bool')
            m[0,0,0,1,nrows-1,0] = True
            m = m*idcol
            if ncols > 1:
                m[(2,0,np.arange(ncols-1),1,nrows-1,np.arange(1,ncols))] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # (0,i,j)->(2,i,j-1) (1,i,j)->(2,i+1,j-1), needs e^ikb2
            m = np.zeros((3,1,ncols,3,1,ncols), dtype='bool')
            m[0,0,0,2,0,ncols-1] = True
            m = m*idrow
            if nrows > 1:
                m[(1,np.arange(nrows-1),0,2,np.arange(1,nrows),ncols-1)] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # (1,i,j)->(2,i+1,j-1) corner, needs e^ik(b2-b1)
            m = np.zeros((3,nrows,ncols,3,nrows,ncols), dtype='bool')
            m[1,nrows-1,0,2,0,ncols-1] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # (2,i,j)->(1,i-1,j+1) corner, needs e^-ik(b2-b1)
            m = np.zeros((3,nrows,ncols,3,nrows,ncols), dtype='bool')
            m[2,0,ncols-1,1,nrows-1,0] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # (2,i,j)->(0,i,j+1), (2,i,j)->(1,i-1,j+1), needs e^-ikb2
            m = np.zeros((3,1,ncols,3,1,ncols), dtype='bool')
            m[2,0,ncols-1,0,0,0] = True
            m = m*idrow
            if nrows > 1:
                m[(2,np.arange(1,nrows),ncols-1,1,np.arange(nrows-1),0)] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # (1,i,j)->(0,i+1,j), (1,i,j)->(2,i+1,j-1), needs e^-ikb1
            m = np.zeros((3,nrows,1,3,nrows,1), dtype='bool')
            m[1,nrows-1,0,0,0,0] = True
            m = m*idcol
            if ncols > 1:
                m[(1,nrows-1,np.arange(1,ncols),2,0,np.arange(ncols-1))] = True
            self.masks.append(np.reshape(m, kinshape, order='F').transpose())
            # Sanity check
            # The masks should not overlap
            assert (sum([a.sum() for a in self.masks]) ==
                    np.sum(self.masks[0] | self.masks[1] | self.masks[2] |
                           self.masks[3] | self.masks[4] | self.masks[5]))
    #
    def set_kinetic_random(self,t,wt=0,we=0):
        """
        Creates a kinetic energy matrix with some random noise in it.

        Random noise is uniform. Noise ofhopping constants centred on
        t and noise of on-site energy centred on 0.

        Inputs: t - number. Hopping constant.
            wt - number. Width of the random noise of hopping.
            we - number. Width of the random noise of on-site energy.
        Effects: sets self.kin

        Last Modified: 2020-11-13
        """
        rng = np.random.default_rng()
        # Get the regular kinetic energy set
        self.set_kinetic(t, force_generic=True)
        self.modified_kin = True
        # Now apply random noise.
        # Get some random noise for the whole matrix.
        noise = rng.random(self.kin.shape)*wt - wt/2
        # Zero out elements which should be zero.
        noise[self.kin == 0] = 0
        # Make the noise matrix Hermitian.
        # Zero out the lower triangle, otherwise next step will be sum of two
        # random variates which is not a uniform distribution.
        noise = np.triu(noise) 
        noise = noise + noise.transpose()
        # Apply noise to the kinetic energy.
        self.kin += noise
        # Now get the noisy diagonal for on-site energy.
        self.kin += np.diag(rng.random(len(self.kin))*we - we/2)
        # Done.



#!/usr/bin/python3

"""
Mean field Hubbard model of Kagome lattice. Gamma point only.

Periodic boundary conditions are assumed.

Created: 2020-07-03
Last Modified: 2021-05-18
Author: Bernard Field
    Copyright (C) 2021 Bernard Field, GNU GPL v3+
"""

import numpy as np
import json
from hubbard.base import Hubbard

class KagomeHubbard(Hubbard):
    #
    ## IO
    #
    def __init__(self,nrows,ncols,u=0,t=1,nup=0,ndown=0,
                 allow_fractions=False,**kwargs):
        """
        Creates a kagome lattice and initialises things.

        nrows and ncols cannot be changed later without building a new instance.

        Inputs: nrows - positive integer, number of rows
                in the supercell.
            ncols - positive integer, number of columns in
                the supercell.
            t - number, optional (default 1). Hopping constant.
            u - number, optional (default 0). Hubbard U constant.
            nup,ndown,kwargs - arguments for set_electrons
                Defaults to no electrons. (Do not set backup)
        Last Modified: 2020-08-10
        """
        # Check that nrows and ncols are positive
        if nrows <= 0:
            raise ValueError("Number of rows must be positive.")
        if ncols <= 0:
            raise ValueError("Number of columns must be positive.")
        # Initialise the important constants.
        self.nrows = nrows
        self.ncols = ncols
        # Call the parent initialiser.
        super().__init__(3*nrows*ncols,u=u,nup=nup,ndown=ndown,
                         allow_fractions=allow_fractions,**kwargs)
        # Initialise the kinetic energy.
        self.set_kinetic(t)
    #
    def copy(self):
        """
        Returns a copy of this kagome lattice object.

        Outputs: KagomeHubbard object.

        Last Modified: 2020-08-04
        """
        # Copy charge density and U.
        kagome = KagomeHubbard(self.nrows,self.ncols,u=self.u,
                               nup=self.nup.copy(),ndown=self.ndown.copy(),
                               allow_fractions=self.allow_fractions)
        # Copy the kinetic energy matrix.
        kagome.set_kinetic_from_matrix(self.kin.copy())
        # Copy magnetic field
        kagome.set_mag(self.mag)
        return kagome
    #
    @classmethod
    def load(cls,f):
        """
        Loads a KagomeHubbard object from a JSON file f.
        Inputs: f - string, filename.
        Outputs: KagomeHubbard object.
        Last Modified: 2020-08-04
        """
        # Load the file
        with open(f) as file:
            di = json.load(file)
        # Create a KagomeHubbard object
        kagome = cls(nrows=di['nrows'],ncols=di['ncols'],
                               u=di['u'],nup=np.asarray(di['nup']),
                               ndown=np.asarray(di['ndown']),
                               allow_fractions=di['allow_fractions'])
        kagome.set_kinetic_from_matrix(np.asarray(di['kin']))
        kagome.set_mag(di['mag'])
        return kagome
    #
    def save(self,f):
        """
        Save a JSON representation of the object's data to file f.
        Inputs: f - string, filename.
        Writes a text file f.
        Last Modified: 2021-05-18
        """
        with open(f,mode='w') as file:
            json.dump({'nrows':int(self.nrows),
                       'ncols':int(self.ncols),
                       'u':float(self.u),
                       'mag':float(self.mag),
                       'nup':self.nup.tolist(),
                       'ndown':self.ndown.tolist(),
                       'allow_fractions':bool(self.allow_fractions),
                       'kin':self.kin.tolist()},
                      file)
            # Have to use tolist because numpy arrays aren't JSON-able.
    #
    ## ELECTRON DENSITY MODIFIERS
    #
    def _electron_density_single_methods(self,nelect,method,up,**kwargs):
        """
        Handler for setting the electron density of the Kagome lattice.
        Accepts the 'star' method and 'points' keyword.
        See also the parent method.
        
        Inputs: nelect - number of electrons.
            method - string, {'star'}, specifies the method.
            up - Boolean. Whether this is spin up electrons.
        Keyword Arguments:
            points - Boolean. Default True. For star method, whether to
                place the spin up electrons at the points of the star.
        Output: electron density - (self.nsites,) ndarray.
        Last Modified: 2020-08-11
        """
        # Process kwargs
        points = kwargs.get('points',True)
        # Check the method.
        if method == 'star':
            # Star method. We alternate based on spin up or down.
            if up:
                density = self._electron_density_star(nelect,points)
            else:
                density = self._electron_density_star(nelect,(not points))
        else:
            # The method is not here. Pass execution up the MRO.
            density = super()._electron_density_single_methods(nelect,
                                                    method,up,**kwargs)
        return density
    #
    def _electron_density_star(self,n,points):
        """
        Produce a star-like configuration of electron density.
        
        I found this to be the ground state of the Kagome Hubbard
        lattice for a good part of the parameter space.

        Inputs: n - integer, number of electrons.
            points - Boolean, whether to put most of the density in
                the points of the stars (True) or the inner bits (False).
        Output: 1D ndarray representing electron density.
        Last Modified: 2020-07-15
        """
        # Get the coordinates of the points of the star.
        rowcolmgrid = np.mgrid[0:self.nrows,0:self.ncols]
        # In the rows and columns, which sublattice site has the star point?
        ncoord = np.mod(rowcolmgrid[0] + 2*rowcolmgrid[1],3)
        # Convert to numerical coordinates.
        pcoords = (ncoord + 3*rowcolmgrid[0] + 3*self.nrows*rowcolmgrid[1]).flatten()
        # A useful constant: number of unit cells
        ncells = self.nrows*self.ncols
        # Four cases, bases on whether to fill points first, and the filling.
        if points:
            if n <= self.nrows*self.ncols: # One third filling or less
                density = np.zeros(3*ncells)
                density[pcoords] = n/ncells
                # Just fill the points of the stars.
            else:
                # Put overflow from star points into other points.
                density = np.ones(3*ncells) * (n/(2*ncells)-1/2)
                density[pcoords] = 1
        else:
            if n <= 2*self.nrows*self.ncols: # Two thirds filling or less
                # Leave the star points empty
                density = np.ones(3*ncells) * (n/(2*ncells))
                density[pcoords] = 0
            else:
                # Put overflow into the star points
                density = np.ones(3*ncells)
                density[pcoords] = n/ncells - 2
        assert abs(density.sum() - n) < 1e-14
        return density
    #
    ## GETTERS
    #
    def get_coordinates(self):
        """
        Return the coordinates for plotting electron density.
        
        Output: a (self.nsites,2) ndarray of floats.
        Last Modified: 2020-08-10
        """
        return kagome_coordinates(self.nrows,self.ncols)
    #
    ## PLOTTERS
    #
    #
    ## SETTERS
    #
    def set_kinetic(self,t):
        """
        Create the kinetic energy part of the kagome Hamiltonian matrix.
        
        Coordinates are sublattice+row*3+col*nrows*3.

        Inputs: t - real number. Hopping constant.
        Effect: sets self.kin to a 3*nrows*ncols dimensional square ndarray.

        Last Modified: 2020-08-10
        """
        # Have the kinetic energy term in tensor form,
        # where sublattice site, row and column are separate indices.
        tensor = -t*kagome_adjacency_tensor(self.nrows,self.ncols)
        # Reshape to a square matrix.
        length = self.nsites
        self.kin = np.reshape(tensor,(length,length),order='F')
    #
    def set_kinetic_random(self,t,wt=0,we=0):
        """
        Create a kinetic energy matrix with some random noise in it.
        
        Random noise is uniform. Noise of hopping constants centred on
        t and noise of on-site energy centred on 0.

        Inputs: t - number. Hopping constant.
            wt - number. Width of the random noise of hopping.
            we - number. Width of the random noise of on-site energy.
        Effects: sets self.kin

        Last Modified: 2020-11-13
        """
        rng = np.random.default_rng()
        # Get the regular kinetic energy set.
        self.set_kinetic(t)
        # Get some random noise for the whole matrix.
        noise = rng.random(self.kin.shape)*wt - wt/2
        # Zero out elements which should be zero.
        noise[self.kin == 0] = 0
        # Make the noise matrix Hermitian.
        # Zero out the lower triangle, otherwise next step will be sum of two random
        # variates which is not a uniform distribution.
        noise = np.triu(noise) 
        noise = noise + noise.transpose()
        # Apply noise to the kinetic energy.
        self.kin += noise
        # Now get the noisy diagonal for on-site energy.
        self.kin += np.diag(rng.random(len(self.kin))*we - we/2)
        # Done.
    #


def simulate(nrows,ncols,nelectup,nelectdown,u,**kwargs):
    """
    Sets up and does a Hubbard calculation on a kagome lattice.

    I recommend you look to hubbard.utils instead. This is deprecated.

    Inputs: nrows - positive integer. Number of rows in the supercell.
        ncols - positive integer. Number of columns in the supercell.
        nelectup - non-negative integer. Number of spin up electrons.
            Must be no greater than 3*nrows*ncols.
        nelectdown - non-negative integer. Number of spin down electrons.
            Must be no greater than 3*nrows*ncols.
        u - real number. Hubbard U parameter.
    Keyword Arguments:
        t - real number, default 1. Hopping parameter.
        scheme - string, default 'linear'. Mixing scheme.
            Currently, only linear mixing is implemented.
        mix - number between 0 and 1, default 0.5. Linear mixing parameter.
        ediff - positive number, default 1e-6. Energy difference below which
            we assume the self-consistency cycle has converged.
        max_iter - positive integer, default 500. Maximum number of iterations
            of self-consistency to perform.
        initial - string, default 'random'. Determines how the electron density
            is initialised. Options are 'random' and 'uniform'.
        alpha - optional positive number, default None. For 'random' initial.
            Dirichlet parameter for random generation. If not set, is
            automatically chosen for optimal spread.

    Outputs: KagomeHubbard object.
    
    Last Modified: 2020-07-10
    """
    # Process default values. kwargs.get(key,default)
    # t, hopping parameter.
    t = kwargs.get('t',1)
    # Mixing scheme.
    scheme = kwargs.get('scheme','linear')
    # Parameters for mixing scheme.
    if scheme == 'linear':
        mix = kwargs.get('mix',0.5)
        ediff = kwargs.get('ediff',1e-6)
        max_iter = kwargs.get('max_iter',500)
    else:
        raise ValueError("Mixing scheme "+str(scheme)+" is not implemented.")
    # Density initialisation scheme
    initial = kwargs.get('initial','random')
    if initial == 'random':
        alpha = kwargs.get('alpha',None)
    # Input processing complete.
    # Create the object
    kagome = KagomeHubbard(nrows,ncols,t=t,u=u,nup=nelectup,ndown=nelectdown,
                           method=initial,alpha=alpha)
    # Run the simulation
    if scheme == 'linear':
        kagome.linear_mixing(max_iter,ediff,mix)
    else:
        raise ValueError("Mixing scheme "+str(scheme)+" is not implemented.")
    return kagome

def kagome_coordinates(nrows,ncols):
    """
    For each point in the kagome lattice, gives its Cartesian coordinates.

    Inputs: nrows and ncols, integers, number of rows and columns in supercell.
    Output: a (3*nrows*ncols, 2) array
    Last Modified: 2020-08-10
    """
    # Lattice vectors.
    a1 = np.array([1/2,np.sqrt(3)/2])
    a2 = np.array([-1/2,np.sqrt(3)/2])
    # Sub-lattice vectors
    sub = np.array([[0,0],a1/2,a2/2])
    # There's probably a vectorised way to do this, but I don't know it.
    # I generally don't need to run this too frequently, and its
    # efficiency isn't atrocious.
    coords = np.empty((3,nrows,ncols,2))
    for i in range(3):
        for j in range(nrows):
            for k in range(ncols):
                coords[i,j,k] = sub[i]+j*a1+k*a2
    return np.reshape(coords,(3*nrows*ncols,2),order='F')

def kagome_adjacency_tensor(nrows, ncols):
    """
    Creates the adjacency tensor for the kagome lattice.
    
    First three coords is input. Second 3 is output.
    Returns a 3*nrows*ncols*3*nrows*ncols ndarray.
    Elements are 1 or 0. Can also have 2's if nrows or ncols is 1.
    Last Modified: 2020-06-03
    """
    # Initialise the adjacency tensor.
    adj = np.zeros((3,nrows,ncols,3,nrows,ncols))
    # We have 12 adjacencies (counting each direction separately)
    # Use multiplication and broadcasting.
    #(0,i,j)->(1,i,j)
    # Generate matrices mapping each of coordinates.
    m0to1 = np.zeros((3,1,1,3,1,1))
    m0to1[0,0,0,1,0,0] = 1
    idrow = np.identity(nrows).reshape(1,nrows,1,1,nrows,1)
    idcol = np.identity(ncols).reshape(1,1,ncols,1,1,ncols)
    # Multiply together to get the full operation.
    op = m0to1*idrow*idcol
    # Add to the adjacency matrix.
    adj += op
    # (0,i,j)->(2,i,j)
    m0to2 = np.zeros((3,1,1,3,1,1))
    m0to2[0,0,0,2,0,0] = 1
    op = m0to2*idrow*idcol
    adj += op
    # (0,i,j)->(1,i-1,j)
    rowm1 = np.zeros((nrows,nrows)) # Row minus 1
    rowm1[(np.arange(nrows),np.mod(np.arange(nrows)-1,nrows))] = 1
    # mod implements periodic boundary conditions.
    rowm1 = rowm1.reshape((1,nrows,1,1,nrows,1))
    op = m0to1*rowm1*idcol
    adj += op
    # (0,i,j)->(2,i,j-1)
    colm1 = np.zeros((ncols,ncols))
    colm1[(np.arange(ncols),np.mod(np.arange(ncols)-1,ncols))] = 1
    colm1 = colm1.reshape((1,1,ncols,1,1,ncols))
    op = m0to2*idrow*colm1
    adj += op
    # (1,i,j)->(0,i,j)
    m1to0 = np.zeros((3,1,1,3,1,1))
    m1to0[1,0,0,0,0,0] = 1
    op = m1to0*idrow*idcol
    adj += op
    # (1,i,j)->(2,i,j)
    m1to2 = np.zeros((3,1,1,3,1,1))
    m1to2[1,0,0,2,0,0] = 1
    op = m1to2*idrow*idcol
    adj += op
    # (1,i,j)->(0,i+1,j)
    rowp1 = np.zeros((nrows,nrows)) # Row plus 1
    rowp1[(np.arange(nrows),np.mod(np.arange(nrows)+1,nrows))] = 1
    rowp1 = rowp1.reshape((1,nrows,1,1,nrows,1))
    op = m1to0*rowp1*idcol
    adj += op
    # (1,i,j)->(2,i+1,j-1)
    op = m1to2*rowp1*colm1
    adj += op
    # (2,i,j)->(0,i,j)
    m2to0 = np.zeros((3,1,1,3,1,1))
    m2to0[(2,0,0,0,0,0)] = 1
    op = m2to0*idrow*idcol
    adj += op
    # (2,i,j)->(1,i,j)
    m2to1 = np.zeros((3,1,1,3,1,1))
    m2to1[(2,0,0,1,0,0)] = 1
    op = m2to1*idrow*idcol
    adj += op
    # (2,i,j)->(0,i,j+1)
    colp1 = np.zeros((ncols,ncols))
    colp1[(np.arange(ncols),np.mod(np.arange(ncols)+1,ncols))] = 1
    colp1 = colp1.reshape((1,1,ncols,1,1,ncols))
    op = m2to0*idrow*colp1
    adj += op
    # (2,i,j)->(1,i-1,j+1)
    op = m2to1*rowm1*colp1
    adj += op
    # Sanity checking
    # We have the right number of elements
    assert adj.sum()==12*nrows*ncols
    # The tensor is symmetric
    assert (adj == adj.transpose((3,4,5,0,1,2))).all()
    # Return
    return adj



#!/usr/bin/python3

"""
Base class for the Mean field Hubbard model, at the Gamma point only.

    Copyright (C) 2021 Bernard Field, GNU GPL v3+

Has the mixing routines and other things which aren't dependent on the
specifics of the lattice.

There are a few main differences from the parent class HubbardKPoints.
 - While HubbardKPoints dynamically generates the kinetic energy matrix
    for a given k, here the kinetic energy has no variables which change
    mid-calculation, so the kinetic energy matrix is cached as self.kin.
    Many methods here access self.kin directly. set_kinetic is expected to
    set self.kin, rather than merely record parameters.
 - _eigensystem is streamlined to exploit the single k-point calculation.
 - set_kmesh won't do anything other than make the Gamma point.

A key advantage of the Gamma point only routine is that it simplifies some
algorithms. Building kinetic energy matrices is easier without having to figure
out where momentum comes into it. The move_electrons method currently is only
implemented for Gamma point only (although that's mainly a matter of it not being
the most useful method so not worth development time). Having a cached kinetic energy
matrix rather than needing to remember a bunch of parameters can make complex
Hamiltonians easier to deal with.
There might be a small performance improvement as well, although I haven't tested it.

Several things must be defined in the subclasses.
 - The kinetic energy matrix, self.kin. This needs to be set in the 
     initialisation step. Ideally you'd also allow it to be
     changed later. THIS ONE IS IMPORTANT.
 - The copy method. find_magnetic_minimum won't work if this isn't defined. And
    I don't know which data you need to feed into new instances.
 - get_coordinates - returns a (nsites, 2) ndarray which indicates
     the Cartesian coordinates of each site. Used for plotting.
     (I assume a 2D lattice for plotting. 3D plotting will require re-working
     the plotting routines to do 3D plots. That's a job for a subclass.)
 - save and load methods (you don't have to define them, but they aren't
     defined here). (I may find a way to abstract the saving/loading later.)
 - any bespoke electron density setting methods, in _electron_density_single_methods

Created: 2020-08-10
Last Modified: 2020-09-17
Author: Bernard Field
"""

from warnings import warn

import numpy as np

from hubbard.kpoints.base import HubbardKPoints

class Hubbard(HubbardKPoints):
    """Base class for the Mean field Hubbard model, at the Gamma point only."""
    #
    ## IO
    #
    # Everything of interest for IO is in the parent class.
    # The only thing to look out for is k-points, but that defaults to initialising
    # as Gamma point only so that's okay.
    #
    ## ELECTRON DENSITY MODIFIERS
    #
    def find_magnetic_minimum(self,max_iter=500,ediff=1e-6,mix=0.5,verbose=False):
        """
        Flips the spin of electrons (self-consistently) until we find an
        energy minimum.
        I do not recommend this method any more, although it is useable.
        A better method is linear_mixing with finite T.
        I note that for large U the energy landscape is difficult to
        traverse. This method (and linear_mixing) will only find a
        local minimum.
        
        Inputs: max_iter, ediff, mix - as for linear_mixing.
            verbose - Boolean, default False. Prints progress.
        
        Last Modified: 2020-08-10
        """
        # To save on typing, stick the keyword parameters into a dictionary
        # for linear_mixing.
        linmixparams = {
            'max_iter':max_iter,
            'ediff':ediff,
            'mix':mix,
            'print_residual':False
            }
        # First step: ensure current state is self-consistent
        en = self.linear_mixing(**linmixparams)
        # Decide if we want to go up or down, or stay put.
        # Energy of flipping one from down to up.
        nsites = self.nsites
        if self.nelectdown >= 1 and self.nelectup <= nsites-1:
            # Can we do such a flip?
            if verbose: print("Checking first up spin flip.")
            kagup = self.copy()
            kagup.move_electrons(1,-1)
            enup = kagup.linear_mixing(**linmixparams)
        else:
            # If not, set energy to very large value so we don't cross it.
            enup = 1e15
        # Energy of flipping from up to down
        if self.nelectup >= 1 and self.nelectdown <= nsites-1:
            # Can we do such a flip?
            if verbose: print("Checking first spin down flip.")
            kagdown = self.copy()
            kagdown.move_electrons(-1,1)
            endown = kagdown.linear_mixing(**linmixparams)
        else:
            # If not, set energy to very large value so we don't cross it.
            endown = 1e15
        # Check what we want to do.
        # If current energy is minimum, stay put
        if en <= enup and en <= endown:
            if verbose: print("Already at energy minimum.")
            return
        # Else, take the path of most energy drop.
        if enup < endown:
            go_up = True
        elif enup > endown:
            go_up = False
        else: # We have a tie.
            # Go towards zero magnetization,
            # going up if at zero.
            go_up = self.get_magnetization() <= 0
        # Adjust electrons in that direction.
        if go_up:
            if verbose: print("Searching upwards.")
            self.set_electrons(nup=kagup.nup,ndown=kagup.ndown)
            en = enup
        else:
            if verbose: print("Searching downwards.")
            self.set_electrons(nup=kagdown.nup,ndown=kagdown.ndown)
            en = endown
        #Iterate in that direction until a minimum is found.
        at_edge = lambda : ((go_up and (self.nelectup > nsites-1 and
                                       self.nelectdown < 1))
                            or ((not go_up) and (self.nelectdown > nsites-1 and
                                                 self.nelectup < 1)))
        while not at_edge():
            # Create the lattice with the next spin flip.
            kagnext = self.copy()
            if go_up:
                kagnext.move_electrons(1,-1)
            else:
                kagnext.move_electrons(-1,1)
            if verbose: print("Checking magnetization "+str(self.get_magnetization()))
            ennext = kagnext.linear_mixing(**linmixparams)
            # Check if the energy is higher.
            if ennext >= en:
                # If so, we're done
                break
            # Otherwise, set the current to the trial step
            self.set_electrons(nup=kagnext.nup,ndown=kagnext.ndown)
            en = ennext
        # Done
        if verbose: print("Found energy minimum.")
        return
    #
    def move_electrons(self,up=0,down=0):
        """
        Add or subtract spin up and down electrons, based on eigenstates.

        Because doing this changes the density and thus the Hamiltonian,
        repeated calls to this function will give a different result than
        a single call to this function.

        Inputs: up - integer, default 0. Number of spin up electrons to add.
                If negative, subtracts.
            down - integer, default 0. Number of spin down electrons to add.
                If negative, subtracts.
        Last Modified: 2020-09-16
        """
        # Check that up and down are valid.
        if up+self.nelectup > self.nsites or up+self.nelectup < 0:
            raise ValueError("Trying to put number of up electrons out of bounds.")
        if down+self.nelectdown > self.nsites or down+self.nelectdown < 0:
            raise ValueError("Trying to put number of down electrons out of bounds.")
        # Get the eigenvectors
        _, _, vup, vdown = self._eigensystem()
        # We want the amplitude squared of the eigenvectors
        vup = vup ** 2
        vdown = vdown ** 2
        if not self.allow_fractions:
            # Adjust the spin up density.
            if up > 0:
                self.nup += vup[:,self.nelectup:self.nelectup+up].sum(1)
            elif up < 0:
                self.nup -= vup[:,self.nelectup+up:self.nelectup].sum(1)
            # Adjust the spin down density.
            if down > 0:
                self.ndown += vdown[:,self.nelectdown:
                                    self.nelectdown+down].sum(1)
            elif down < 0:
                self.ndown -= vdown[:,self.nelectdown+down:
                                    self.nelectdown].sum(1)
        else:
            # Adjust the spin up densoity.
            fracN = self.nelectup - int(self.nelectup) # fractional occupation of current
            frac = self.nelectup+up - int(self.nelectup+up)# fractional of final.
            if up > 0:
                # Fractional occupation of current state
                self.nup += vup[:,int(self.nelectup)]*min(1-fracN,up)
                if 1-fracN < up:
                    # Still have some up left over.
                    # Add whole states.
                    self.nup += vup[:,int(self.nelectup)+1:int(self.nelectup+up)].sum(1)
                    # Add the remaining fractional part
                    self.nup += vup[:,int(self.nelectup+up)]*frac
            elif up < 0:
                self.nup -= vup[:,int(self.nelectup)]*min(fracN,-up)
                if fracN < -up:
                    self.nup -= vup[:,int(self.nelectup+up)+1:int(self.nelectup)].sum(1)
                    self.nup -= vup[:,int(self.nelectup+up)] * (1-frac)
            # Adjust the spin down density.
            fracN = self.nelectdown - int(self.nelectdown) # fractional occupation of current
            frac = self.nelectdown+down - int(self.nelectdown+down)# fractional of final.
            if down > 0:
                self.ndown += vdown[:,int(self.nelectdown)]*min(1-fracN,down)
                if 1-fracN < down:
                    self.ndown += vdown[:,int(self.nelectdown)+1:
                                        int(self.nelectdown+down)].sum(1)
                    self.ndown += vdown[:,int(self.nelectdown+down)]*frac
            elif down < 0:
                self.ndown -= vdown[:,int(self.nelectdown)]*min(fracN,-down)
                if fracN < -down:
                    self.ndown -= vdown[:,int(self.nelectdown+down)+1:
                                        int(self.nelectdown)].sum(1)
                    self.ndown -= vdown[:,int(self.nelectdown+down)] * (1-frac)
        # Adjust the numbers of electrons.
        self.nelectup += up
        self.nelectdown += down
    #
    ## GETTERS
    #
    def get_kinetic(self,k=None):
        """
        Returns the kinetic energy matrix.

        Inputs: None (k is a dummy variable for compatibility with parent class.)
        Output: (nsites,nsites) ndarray
        Last Modified: 2020-09-17
        """
        return self.kin
    #
    def _eigensystem(self):
        """
        Solves the eigensystem and returns the eigenenergies and states.

        Simplified due to only having a single k-point.
        
        Outputs: eup, edown, vup, vdown
            eup, edown - (nsites,) ndarray of numbers, sorted. Eigenenergies.
            vup, vdown - (nsites,nsites) ndarray of numbers. Eigenvectors.
                v[:,i] is the normalised eigenvector corresponding to the
                eigenvalue e[i], as output by np.linalg.eigh.

        Last Modified: 2020-09-16
        """
        # Get the potential matrices
        potup, potdown, _ = self._potential()
        # Solve for the single-electron energy levels.
        eup, vup = np.linalg.eigh(self.kin + potup)
        edown, vdown = np.linalg.eigh(self.kin + potdown)
        return eup, edown, vup, vdown
    #
    ## PLOTTERS
    #
    #
    ## SETTERS
    #
    def set_kinetic(self):
        """
        Use this method to set self.kin.
        self.kin is a Hermitian (nsites,nsites) ndarray.
        Implementation is lattice-dependent.
        """
        raise NotImplementedError
    #
    def set_kinetic_from_matrix(self,m):
        """
        Primarily for use when you have saved a matrix for later use.

        Inputs: m - array-like, Hermitian matrix, the Kinetic Energy
            and also all other single-particle terms in the Hamiltonian.
        Effect: sets self.kin

        Last Modified; 2020-08-10
        """
        # Check that m is valid.
        m = np.asarray(m)
        # Check that is it the right shape.
        if m.shape != (self.nsites, self.nsites):
            raise ValueError("m is not of the right shape.")
        # Check that it is Hermitian
        if not np.all(np.abs(m - m.conjugate().transpose()) < 1e-14):
            raise ValueError("m is not Hermitian.")
        # Set the matrix
        self.kin = m
    #
    def set_kmesh(self,*args,**kwargs):
        """
        As a Gamma-point only code, this is does nothing much.

        Sets kpoints to 1 and kmesh to be just the Gamma point.
        Complains if you try to make it do anything else.
        Last Modified: 2020-09-17
        """
        # If setting k-mesh to Gamma point only (as in initialisation), be silent.
        # Also be silent if no arguments provided.
        if not ((not args and not kwargs) or np.array_equal(args,[1]*self.dims) or
                np.array_equal(args,[[0]*self.dims])):
            warn("Attempted to change kmesh in Gamma point only implementation. "
                 "Ignoring.")
        self.kpoints = 1
        self.kmesh = np.array([[0]*self.dims])



#!/usr/bin/python3

"""
Helper scripts for analysing my Mean field Hubbard model of the Kagome lattice.

Created: 2020-08-04
Last Modified: 2021-04-16
Author: Bernard Field
    Copyright (C) 2021 Bernard Field, GNU GPL v3+
"""

import warnings
import os.path
from time import sleep
from glob import glob

import numpy as np

from hubbard.misc import ConvergenceWarning, MixingError, choose_alpha
from hubbard.kagome import KagomeHubbard
import hubbard.progress as progress


def converge(hub,rdiff,rdiff_initial=1e-2,T=None,debug=False,interval=500,mu=None):
    """
    Brings a Hubbard model to self-consistency,
    regardless of how long it takes.
    Inputs: hub - Hubbard object.
        rdiff - number, target residual
        rdiff_initial - number, target residual for the initial linear mixing step.
        T - number, optional. Temperature for Fermi Dirac distribution.
        debug - Boolean, whether to print progress.
        interval - integer. How often to print progress.
        mu - number, optional. Chemical potential for GCE mixing.
    Last Modified: 2021-04-16
    """
    # Suppress ConvergenceWarnings.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore",ConvergenceWarning)
        # We need to perform an initial linear mixing.
        res = hub.residual(T,mu) # Initial determination of the residual.
        while res > rdiff_initial:
            last_residual = res
            res = hub.linear_mixing(max_iter=interval,ediff=0.1,
                                 rdiff=rdiff_initial,T=T,print_residual=False,mu=mu,return_energy=False)
            if debug: print("Linear mixing. Residual={0}".format(res))
            # Catch a pathological case where residual doesn't decrease.
            if np.isclose(last_residual,res):
                if debug: print("Residual didn't decrease. Trying smaller mixing.")
                res = hub.linear_mixing(max_iter=interval,ediff=0.1,mix=0.1,
                                 rdiff=rdiff_initial,T=T,print_residual=False,mu=mu,return_energy=False)
                if debug: print("Linear mixing (mix=0.1). Residual={0}".format(res))
        # Now do main mixing.
        while res > rdiff:
            try:
                res = hub.pulay_mixing(max_iter=interval,rdiff=rdiff,T=T,mu=mu)
                if debug: print("Pulay mixing. Residual={0}".format(res))
            except MixingError:
                # Pulay mixing has a chance to take the electron density out of
                # bounds. It can happen when Pulay mixing diverges.
                # Since there was an error, the residual was not returned.
                if debug:
                    print("Failure in Pulay mixing. Residual={0}. "
                          "Doing some linear mixing.".format(hub.residual(T,mu)))
                res = hub.linear_mixing(max_iter=10,rdiff=rdiff,ediff=0.1,T=T,mu=mu, return_energy=False)
                if debug: print("Linear mixing. Residual={0}".format(res))

def sweep_spin(template,n,nsteps,rdiff,rdiff_initial=1e-2,
               T=None,interval=500,positive_only=True,repeats=1,verbose=False,
               mmin=None,mmax=None,quiet=False,mu=None):
    """
    For a given set of parameters, steps through different nup and ndown ratios.
    Relaxes the Hubbard systems, and returns all the results as a list.
    Has special handling of KeyboardInterrupt. Interrupting once will skip the
    current data point. Interrupting twice within 0.5 seconds will abort the sweep,
    but still return the data collected so far.
    
    Inputs: template - template Hubbard object (with right size,
            U, kinetic energy, etc., but electrons are ignored)
        n - number of electrons.
        nsteps - how many steps (as in linspace).
        rdiff, rdiff_initial, T, interval - as for converge
        positive_only - Boolean. Only check positive magnetization.
            Recommend True when magnetic field is zero, becasue symmetry.
            Ignored if mmin is set.
        repeats - positive integer. How many times to repeat each measurement.
        verbose - Boolean.
        mmin - optional number. Minimum magnetization to look at.
            Defaults to 0 if positive_only else -mmax.
        mmax - optional number. Maximum magnetization to look at.
            Defaults to the smaller of n or nsites.
        quiet - Boolean. Verbose overrules.
        mu - optional number. Chemical potential for GCE mixing.
    
    Output: list of Hubbard objects.

    Last Modified: 2021-04-26
    """
    # Initialise
    hub_list = [] # Results list
    nsites = template.nsites
    # Bounds
    if mmax is None:
        mmax = min(n,nsites)
    if mmin is None:
        if positive_only:
            mmin = 0
        else:
            mmin = -mmax
    nupmin = max(0,n-nsites,(n+mmin)/2)
    nupmax = min((mmax+n)/2,n,nsites)
    # Check an error case
    if nupmin != nupmax and nsteps == 1:
        raise ValueError("Cannot have nsteps=1 when have unequal mmin and mmax.")
    # Pick an alpha
    alpha = choose_alpha(nsites,min(nupmax,nsites/2))*10
    # Progress bar
    show_progress = (not verbose) and (not quiet)
    if show_progress:
        if nupmin == nupmax:
            progress.new(60,repeats,0)
        else:
            progress.new(60,nupmax,nupmin)
    elif verbose:
        print("Nup from "+str(nupmin)+" to "+str(nupmax))
    # Iterate
    samples,stepsize = np.linspace(nupmin,nupmax,nsteps,retstep=True)
    try:
        for nup in samples:
            for i in range(repeats):
                try:
                    if verbose: print("Nup="+str(nup)+", repeat "+str(i+1))
                    # Get a copy I can work with.
                    hub = template.copy()
                    # Set the electron density
                    hub.set_electrons(nup=nup,ndown=n-nup,method='random',alpha=alpha)
                    # Converge
                    converge(hub,rdiff,rdiff_initial,T=T,interval=interval,debug=verbose,mu=mu)
                    # Record
                    hub_list.append(hub)
                except KeyboardInterrupt:
                    # If interrupt, give a half second to wait for a second interrupt.
                    if verbose: print("Skipping...")
                    sleep(0.5)
                if show_progress:
                    if nupmin==nupmax: progress.update(i+1)
                    else: progress.update(nup+i/repeats*stepsize)
    except KeyboardInterrupt:
        # If we interrupt, still return cleanly.
        if show_progress:
            print('X',end='')
        elif verbose:
            print("Aborting.")
    if show_progress:
        progress.end()
    elif verbose:
        print("Finished.")
    # Return
    return hub_list


def write(kagome,directory='.',ext=''):
    """
    Write kagome to a file in directory.
    Automatically handles naming conventions.
    Filename includes U, supercell, and number of electrons.
    Other details should be specified in ext.
    Inputs: kagome - KagomeHubbard object.
        directory - string to a directory.
        ext - string. Extension to append to automatically
            generated name (but before the number)
    Last Modified: 2020-08-04
    """
    # Check existence of directory.
    if not os.path.isdir(directory):
        raise FileNotFoundError(str(directory)+" not found.")
    # Create the base name
    base_name = ('U'+str(kagome.u)+'cell'+str(kagome.nrows)+'x'+str(kagome.ncols)
            +'n'+str(int(round(kagome.get_electron_number())))+ext)
    # Append number.
    i = 0
    true_name = base_name + '_' + str(i).zfill(4)
    # Check existence.
    while os.path.isfile(directory+'/'+true_name):
        # If exists, try the next number.
        i += 1
        true_name = base_name + '_' + str(i).zfill(4)
    # Write
    kagome.save(directory+'/'+true_name)

def write_list(kagome_list,directory='.',ext=''):
    """
    Iterates over kagome_list and applies write() to each element.
    Last Modified: 2020-08-06
    """
    for kagome in kagome_list:
        write(kagome,directory,ext)

def read_list(directory='.', pattern='U*cell*x*n*', cls=KagomeHubbard):
    """
    Reads a directory filled with Hubbard files.

    Uses glob matching for the pattern.

    Inputs: directory - string to a directory
        pattern - string, glob pattern for the files to load.
        cls - class. Class to interpret the files as.
    Output: list of cls objects.
    Last Modified: 2020-09-18
    """
    true_pattern = directory + '/' + pattern
    files = glob(true_pattern)
    kag_list = [ cls.load(f) for f in files ]
    return kag_list

def boltzmann_average(hub_list,T,func,Ten=None):
    """
    Takes the mean of some value of Hubbard objects,
    weighted by a Boltzmann factor.
    Can also take the minimum energy value with T=0.

    Inputs: hub_list - list of Hubbard objects
        T - number, temperature for the Boltzmann factor.
        func - callable. Takes a KagomeHubbard object and returns a number.
        Ten - number, optional. Temperature for calculating the energy.
    Output: a number
    Last Modified: 2020-08-11
    """
    # Energies
    en = [h.energy(Ten) for h in hub_list]
    if T == 0:
        # Filter for minimum energy states.
        boltz = np.zeros_like(en)
        boltz[en == min(en)] = 1
        boltz /= boltz.sum() # Normalise
    else:
        # Weight states by a Boltzmann factor.
        boltz = np.exp(-(np.asarray(en)-min(en))/T)
        # Subtract minimum energy to avoid overflow.
        # Does not change the outcome.
        boltz /= boltz.sum() # Normalise
    # Evaluate the function
    vals = np.array([ func(h) for h in hub_list ])
    # Return, taking the weighted average.
    return np.sum(vals*boltz)

def cull_list(hub_list,e,T=None, mu=None):
    """
    Returns a hub_list with only states with energy less than e above
    the minimum energy.
    Inputs: hub_list - list of Hubbard objects
        e - number
        T - number, optional, temperature for calculating energy.
        mu - number, optional, chemical potential for energy.
    Output: list of Hubbard objects.
    Last Modified: 2021-04-14
    """
    en = np.array([h.energy(T,mu) for h in hub_list])
    return list(np.asarray(hub_list)[en <= (en.min()+e)])
